# coding=utf-8
from __future__ import unicode_literals

import hashlib
import os
import platform
import random
import sys

try:
    import ConfigParser as configparser
except ImportError:
    import configparser

# download delegate
__wget__ = ('WGET_PATH',)
__thunder__ = ('XUNLEI_LX_PATH',)
__transmission__ = ('TRANSMISSION_RPC_URL', 'TRANSMISSION_RPC_PORT',)
__aria2__ = ('ARIA2_RPC_URL', 'ARIA2_RPC_TOKEN',)

__download_delegate__ = __wget__ + __thunder__ + __aria2__ + __transmission__

# fake __all__
__all__ = ('BANGUMI_MOE_URL', 'SAVE_PATH', 'DOWNLOAD_DELEGATE',
           'MAX_PAGE', 'DATA_SOURCE', 'TMP_PATH', 'DANMAKU_API_URL',
           'LANG', 'FRONT_STATIC_PATH', 'ADMIN_TOKEN', 'SHARE_DMHY_URL',
           'GLOBAL_FILTER', 'ENABLE_GLOBAL_FILTER',
           'TORNADO_SERVE_STATIC_FILES',
           )

# cannot be rewrite
__readonly__ = ('BGMI_PATH', 'DB_PATH', 'CONFIG_FILE_PATH', 'TOOLS_PATH',
                'SCRIPT_PATH', 'SCRIPT_DB_PATH', 'FRONT_STATIC_PATH',)

# writeable
__writeable__ = tuple([i for i in __all__ if i not in __readonly__])

# the real __all__
__all__ = __all__ + __download_delegate__ + __readonly__

DOWNLOAD_DELEGATE_MAP = {
    'rr!': __wget__,
    'aria2-rpc': __aria2__,
    'xunlei': __thunder__,
    'transmission-rpc': __transmission__,
}

if not os.environ.get('BGMI_PATH'):  # pragma: no cover
    if platform.system() == 'Windows':
        BGMI_PATH = os.path.join(os.environ.get('USERPROFILE', None), '.bgmi')
        if not BGMI_PATH:
            raise SystemExit
    else:
        BGMI_PATH = os.path.join(os.environ.get('HOME', '/tmp'), '.bgmi')
else:
    BGMI_PATH = os.environ.get('BGMI_PATH')

DB_PATH = os.path.join(BGMI_PATH, 'bangumi.db')
CONFIG_FILE_PATH = os.path.join(BGMI_PATH, 'bgmi.cfg')

SCRIPT_DB_PATH = os.path.join(BGMI_PATH, 'script.db')
SCRIPT_PATH = os.path.join(BGMI_PATH, 'scripts')
TOOLS_PATH = os.path.join(BGMI_PATH, 'tools')


def read_config():
    c = configparser.ConfigParser()
    if not os.path.exists(CONFIG_FILE_PATH):
        write_default_config()
        return
    c.read(CONFIG_FILE_PATH)

    for i in __writeable__:
        if c.has_option('bgmi', i):
            globals().update({i: c.get('bgmi', i)})

    for i in DOWNLOAD_DELEGATE_MAP.get(DOWNLOAD_DELEGATE, []):
        if c.has_option(DOWNLOAD_DELEGATE, i):
            globals().update({i: c.get(DOWNLOAD_DELEGATE, i)})


def print_config():
    c = configparser.ConfigParser()
    if not os.path.exists(CONFIG_FILE_PATH):
        return

    c.read(CONFIG_FILE_PATH)
    string = ''
    string += '[bgmi]\n'
    for i in __writeable__:
        string += '{0}={1}\n'.format(i, c.get('bgmi', i))

    string += '\n[{0}]\n'.format(DOWNLOAD_DELEGATE)
    for i in DOWNLOAD_DELEGATE_MAP.get(DOWNLOAD_DELEGATE, []):
        string += '{0}={1}\n'.format(i, c.get(DOWNLOAD_DELEGATE, i))
    return string


def write_default_config():
    c = configparser.ConfigParser()
    if not c.has_section('bgmi'):
        c.add_section('bgmi')

    for k in __writeable__:
        v = globals().get(k, '0')
        if k == 'ADMIN_TOKEN' and v is None:
            if sys.version_info > (3, 0):
                v = hashlib.md5(str(random.random()).encode('utf-8')).hexdigest()
            else:
                v = hashlib.md5(str(random.random())).hexdigest()

        c.set('bgmi', k, v)

    if DOWNLOAD_DELEGATE not in DOWNLOAD_DELEGATE_MAP.keys():
        raise Exception(DOWNLOAD_DELEGATE)

    if not c.has_section(DOWNLOAD_DELEGATE):
        c.add_section(DOWNLOAD_DELEGATE)

    for k in DOWNLOAD_DELEGATE_MAP.get(DOWNLOAD_DELEGATE, []):
        v = globals().get(k, None)
        c.set(DOWNLOAD_DELEGATE, k, v)

    try:
        with open(CONFIG_FILE_PATH, 'w') as f:
            c.write(f)
    except IOError:
        print('[-] Error writing to config file and ignored')


def write_config(config=None, value=None):
    if not os.path.exists(CONFIG_FILE_PATH):
        write_default_config()
        return {'status': 'error',
                'message': 'Config file does not exists, writing default config file',
                'data': []}

    c = configparser.ConfigParser()
    c.read(CONFIG_FILE_PATH)
    if config is not None and config not in __all_writable_now__:
        result = {'status': 'error',
                  'message': '{0} does not exist or not writeable'.format(config)}
        # return result

    try:
        if config is None:
            result = {'status': 'info',
                      'message': print_config()}

        elif value is None:  # config(config, None)
            result = {'status': 'info'}

            if config in __download_delegate__:
                result['message'] = '{0}={1}'.format(config, c.get(DOWNLOAD_DELEGATE, config))
            else:
                result['message'] = '{0}={1}'.format(config, c.get('bgmi', config))

        else:  # config(config, Value)
            if config in __writeable__:
                if config == 'DOWNLOAD_DELEGATE' and value not in DOWNLOAD_DELEGATE_MAP:
                    result = {'status': 'error',
                              'message': '{0} is not a support download_delegate'.format(value)}
                else:
                    c.set('bgmi', config, value)
                    with open(CONFIG_FILE_PATH, 'w') as f:
                        c.write(f)
                    read_config()
                    if config == 'DOWNLOAD_DELEGATE':
                        if not c.has_section(DOWNLOAD_DELEGATE):
                            c.add_section(DOWNLOAD_DELEGATE)
                            for i in DOWNLOAD_DELEGATE_MAP[DOWNLOAD_DELEGATE]:
                                v = globals().get(i, '')
                                c.set(DOWNLOAD_DELEGATE, i, v)

                            with open(CONFIG_FILE_PATH, 'w') as f:
                                c.write(f)
                    result = {'status': 'success',
                              'message': '{0} has been set to {1}'.format(config, value)}

            elif config in DOWNLOAD_DELEGATE_MAP.get(DOWNLOAD_DELEGATE):
                c.set(DOWNLOAD_DELEGATE, config, value)
                with open(CONFIG_FILE_PATH, 'w') as f:
                    c.write(f)
                result = {'status': 'success',
                          'message': '{0} has been set to {1}'.format(config, value)}
            else:
                result = {'status': 'error',
                          'message': '{0} does not exist or not writeable'.format(config)}

    except configparser.NoOptionError:
        write_default_config()
        result = {'status': 'error', 'message': 'Error in config file, try rerun `bgmi config`'}

    result['data'] = [{'writable': True, 'name': x, 'value': globals()[x]} for x in __writeable__] + \
                     [{'writable': False, 'name': x, 'value': globals()[x]} for x in __readonly__]
    return result


# --------- Writeable ---------- #
# Setting bangumi.moe url
BANGUMI_MOE_URL = 'https://bangumi.moe'

# Setting share.dmhy.org url
SHARE_DMHY_URL = 'https://share.dmhy.org'

# Setting bangumi.moe url
DATA_SOURCE = 'bangumi_moe'

# BGmi user path
SAVE_PATH = os.path.join(BGMI_PATH, 'bangumi')
FRONT_STATIC_PATH = os.path.join(BGMI_PATH, 'front_static')

# admin token
ADMIN_TOKEN = None

# Xunlei offline download
XUNLEI_LX_PATH = os.path.join(BGMI_PATH, 'bgmi-lx')

# temp path
TMP_PATH = os.path.join(BGMI_PATH, 'tmp')

# Download delegate
DOWNLOAD_DELEGATE = 'aria2-rpc'

# danmaku api url, https://github.com/DIYgod/DPlayer#related-projects
DANMAKU_API_URL = ''

# language
LANG = 'zh_cn'

# max page
MAX_PAGE = '3'

# aria2
ARIA2_RPC_URL = 'http://localhost:6800/rpc'
ARIA2_RPC_TOKEN = 'token:'

# path of wget
WGET_PATH = '/usr/bin/wget'

# transmission-rpc
TRANSMISSION_RPC_URL = '127.0.0.1'
TRANSMISSION_RPC_PORT = '9091'

# tag of bangumi on bangumi.moe
BANGUMI_TAG = '549ef207fe682f7549f1ea90'

# Global blocked keyword
GLOBAL_FILTER = 'Leopard-Raws, hevc, x265, c-a Raws'

# enable global filter
ENABLE_GLOBAL_FILTER = '1'

# use tornado serving video files
TORNADO_SERVE_STATIC_FILES = '0'

# ------------------------------ #
# !!! Read config from file and write to globals() !!!
read_config()
# ------------------------------ #
# will be used in other other models
__all_writable_now__ = __writeable__ + DOWNLOAD_DELEGATE_MAP[DOWNLOAD_DELEGATE]

# --------- Read-Only ---------- #
# Python version
IS_PYTHON3 = sys.version_info > (3, 0)

# Detail URL
# platform
IS_WINDOWS = platform.system() == 'Windows'

# - Unify python2 and python3 - #
import codecs
import locale

# Wrap sys.stdout into a StreamWriter to allow writing unicode.

if IS_PYTHON3:
    unicode = str
    if platform.system() != 'Windows':
        file_ = sys.stdout.buffer
        sys.stdout = codecs.getwriter(locale.getpreferredencoding())(file_)
else:
    sys.stdout = codecs.getwriter('utf-8')(sys.stdout)
    input = raw_input


def unicode_(s):
    if not IS_PYTHON3:
        unicode_string = s.decode(sys.getfilesystemencoding())
        return unicode_string
    else:
        return unicode(s)



from __future__ import print_function, unicode_literals
import bgmi.config
from bgmi.downloader.base import BaseDownloadService
from bgmi.config import ARIA2_RPC_URL, ARIA2_RPC_TOKEN
from bgmi.utils import print_info, print_warning, print_success, print_error
from bgmi.lib.models import STATUS_DOWNLOADED, STATUS_NOT_DOWNLOAD, STATUS_DOWNLOADING


if bgmi.config.IS_PYTHON3:
    from xmlrpc.client import ServerProxy, _Method
else:
    from xmlrpclib import ServerProxy, _Method


class _PatchedMethod(_Method):
    def __getitem__(self, name):
        return _Method.__getattr__(self, name)

    def __getattr__(self, name):
        if name == '__getitem__':
            return self.__getitem__
        return _Method.__getattr__(self, name)

    def __call__(self, *args):
        print(args)


class PatchedServerProxy(ServerProxy):
    def __request(self, methodname, params):
        return ServerProxy._ServerProxy__request(self, methodname, params)

    def __getattr__(self, name):
        return _PatchedMethod(self.__request, name)


class Aria2DownloadRPC(BaseDownloadService):
    old_version = False

    def __init__(self, *args, **kwargs):
        self.server = PatchedServerProxy(ARIA2_RPC_URL)
        Aria2DownloadRPC.check_aria2c_version()
        super(Aria2DownloadRPC, self).__init__(**kwargs)

    def download(self):
        if self.old_version:
            self.server.aria2.addUri([self.torrent], {"dir": self.save_path})
        else:
            self.server.aria2.addUri(ARIA2_RPC_TOKEN, [self.torrent], {"dir": self.save_path})
        print_info('Add torrent into the download queue, the file will be saved at {0}'.format(self.save_path))

    @staticmethod
    def check_aria2c_version():
        url = ARIA2_RPC_URL.split('/')
        url[2] = ARIA2_RPC_TOKEN + '@' + url[2]
        url = '/'.join(url)
        s = ServerProxy(url)
        r = s.aria2.getVersion(ARIA2_RPC_TOKEN, )
        version = r['version']
        if version:
            Aria2DownloadRPC.old_version = version < '1.18.4'
        else:
            print_warning('Get aria2c version failed')

    @staticmethod
    def install():
        print_warning('Please install aria2 by yourself')

    def check_download(self, name):
        pass

    @staticmethod
    def download_status(status=None):
        Aria2DownloadRPC.check_aria2c_version()

        print_info('Print download status in database')
        BaseDownloadService.download_status(status=status)
        print()
        print_info('Print download status in aria2c-rpc')
        try:
            server = PatchedServerProxy(ARIA2_RPC_URL)
            # self.server.aria2
            status_dict = {
                STATUS_DOWNLOADING: ['tellActive'],
                STATUS_NOT_DOWNLOAD: ['tellWaiting'],
                STATUS_DOWNLOADED: ['tellStopped'],
                None: ['tellStopped', 'tellWaiting', 'tellActive'],
            }
            for method in status_dict.get(status):
                if method not in ('tellActive', ):
                    params = (0, 1000)
                else:
                    params = ()
                if Aria2DownloadRPC.old_version:
                    data = server.aria2[method](*params)
                else:
                    data = server.aria2[method](ARIA2_RPC_TOKEN, *params)

                if data:
                    print_warning('RPC {0}:'.format(method), indicator=False)

                for row in data:
                    print_success('- {0}'.format(row['dir']), indicator=False)
                    for file_ in row['files']:
                        print_info('    * {0}'.format(file_['path']), indicator=False)

        except Exception as e:
            print_error('Cannot connect to aria2-rpc server')



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import os
import unittest

from bgmi.lib.controllers import *
from bgmi.main import setup
from bgmi.lib.constants import unicode_


class ControllersTest(unittest.TestCase):
    def setUp(self):
        self.bangumi_name_1 = unicode_(os.environ.get('BANGUMI_1'))
        self.bangumi_name_2 = unicode_(os.environ.get('BANGUMI_2'))
        pass

    def test_a_cal(self):
        r = cal()
        self.assertIsInstance(r, dict)
        for day in r.keys():
            self.assertIn(day.lower(), [x.lower() for x in Bangumi.week])
            self.assertIsInstance(r[day], list)
            for bangumi in r[day]:
                self.assertIn("status", bangumi)
                self.assertIn("subtitle_group", bangumi)
                self.assertIn("name", bangumi)
                self.assertIn("keyword", bangumi)
                self.assertIn("update_time", bangumi)
                self.assertIn("cover", bangumi)

    def test_b_add(self):
        r = add(self.bangumi_name_1, 0)
        self.assertEqual(r['status'], 'success')
        r = add(self.bangumi_name_1, 0)
        self.assertEqual(r['status'], 'warning')
        r = delete(self.bangumi_name_1)
        self.assertEqual(r['status'], 'warning')

    def test_c_mark(self):
        r = add(self.bangumi_name_1, 0)
        self.assertEqual(r['status'], 'success')

        r = mark(self.bangumi_name_1, 1)
        self.assertEqual(r['status'], 'success')
        r = mark(self.bangumi_name_1, None)
        self.assertEqual(r['status'], 'info')
        r = mark(self.bangumi_name_2, 0)
        self.assertEqual(r['status'], 'error')

    def test_d_delete(self):
        r = delete()
        self.assertEqual(r['status'], 'warning')
        r = delete(self.bangumi_name_1)
        self.assertEqual(r['status'], 'warning')
        r = delete(self.bangumi_name_1)
        self.assertEqual(r['status'], 'warning')
        r = delete(self.bangumi_name_2)
        self.assertEqual(r['status'], 'error')
        r = delete(clear_all=True, batch=True)
        self.assertEqual(r['status'], 'warning')

    def test_e_search(self):
        r = search(self.bangumi_name_1, dupe=False)

    @staticmethod
    def setUpClass():
        setup()
        recreate_source_relatively_table()



# coding=utf-8
from __future__ import unicode_literals, print_function

import os
import subprocess

from bgmi.lib.models import STATUS_DOWNLOADED, STATUS_NOT_DOWNLOAD, STATUS_DOWNLOADING, Download
from bgmi.utils import print_warning, print_info, print_success


class BaseDownloadService(object):
    def __init__(self, download_obj, save_path, overwrite=True):
        self.name = download_obj.name
        self.torrent = download_obj.download
        self.overwrite = overwrite
        self.save_path = save_path
        self.episode = download_obj.episode
        self.return_code = 0

    def download(self):
        # download
        raise NotImplementedError

    @staticmethod
    def install():
        # install requirement
        raise NotImplementedError

    def check_path(self):
        if not os.path.exists(self.save_path):
            print_warning('Create dir {0}'.format(self.save_path))
            os.makedirs(self.save_path)

    def check_delegate_bin_exist(self, path):
        if not os.path.exists(path):
            raise Exception('{0} not exist, please run command \'bgmi install\' to install'.format(path))

    def call(self, command):
        self.return_code = subprocess.call(command, env={'PATH': '/usr/local/bin:/usr/bin:/bin',
                                                         'HOME': os.environ.get('HOME', '/tmp')})

    def check_download(self, name):
        if not os.path.exists(self.save_path) or self.return_code != 0:
            raise Exception('It seems the bangumi {0} not be downloaded'.format(name))

    @staticmethod
    def download_status(status=None):
        last_status = -1
        for download_data in Download.get_all_downloads(status=status):
            latest_status = download_data['status']
            name = '  {0}. <{1}: {2}>'.format(download_data['id'], download_data['name'],
                                              download_data['episode'])
            if latest_status != last_status:
                if latest_status == STATUS_DOWNLOADING:
                    print('Downloading items:')
                elif latest_status == STATUS_NOT_DOWNLOAD:
                    print('Not downloaded items:')
                elif latest_status == STATUS_DOWNLOADED:
                    print('Downloaded items:')

            if download_data['status'] == STATUS_NOT_DOWNLOAD:
                print_info(name, indicator=False)
            elif download_data['status'] == STATUS_DOWNLOADING:
                print_warning(name, indicator=False)
            elif download_data['status'] == STATUS_DOWNLOADED:
                print_success(name, indicator=False)
            last_status = download_data['status']



# coding: utf-8
from __future__ import print_function, unicode_literals

import os

from bgmi.config import SAVE_PATH, FRONT_STATIC_PATH
from bgmi.front.base import BaseHandler, COVER_URL
from bgmi.lib.models import STATUS_DELETED, STATUS_UPDATING, STATUS_END, Followed
from bgmi.utils import normalize_path


def get_player(bangumi_name):
    episode_list = {}
    bangumi_path = os.path.join(SAVE_PATH, bangumi_name)
    for root, _, files in os.walk(bangumi_path):
        _ = root.replace(bangumi_path, '').split(os.path.sep)
        base_path = root.replace(SAVE_PATH, '')
        if len(_) >= 2:
            episode_path = root.replace(os.path.join(SAVE_PATH, bangumi_name), '')
            if episode_path.split(os.path.sep)[1].isdigit():
                episode = int(episode_path.split(os.path.sep)[1])
            else:
                continue
        else:
            episode = -1

        for bangumi in files:
            if any([bangumi.lower().endswith(x) for x in ['.mp4', '.mkv', '.webm']]):
                video_file_path = os.path.join(base_path, bangumi)
                video_file_path = os.path.join(os.path.dirname(video_file_path), os.path.basename(video_file_path))
                video_file_path = video_file_path.replace(os.path.sep, '/')
                episode_list[episode] = {'path': video_file_path}
                break

    return episode_list


class IndexHandler(BaseHandler):
    def get(self, path):
        if not os.path.exists(FRONT_STATIC_PATH):
            msg = '''<h1>Thanks for your using BGmi</h1>
            <p>It seems you have not install BGmi Frontend, please run <code>bgmi install</code> to install.</p>
            '''
        else:
            msg = '''<h1>Thanks for your using BGmi</h1>
            <p>If use want to use Tornado to serve static files, please run 
            <code>bgmi config TORNADO_SERVE_STATIC_FILES 1</code>, and do not forget install bgmi-frontend by
            running <code>bgmi install</code></p>'''

        self.write(msg)
        self.finish()


class BangumiListHandler(BaseHandler):
    def get(self, type_=''):
        data = Followed.get_all_followed(STATUS_DELETED, STATUS_UPDATING if not type_ == 'old' else STATUS_END)

        if type_ == 'index':
            data.extend(self.patch_list)
            data.sort(key=lambda _: _['updated_time'] if _['updated_time'] else 1)

        for bangumi in data:
            bangumi['cover'] = '{}/{}'.format(COVER_URL, normalize_path(bangumi['cover']))

        data.reverse()

        for item in data:
            item['player'] = get_player(item['bangumi_name'])

        self.write(self.jsonify(data))
        self.finish()



# coding=utf-8
from __future__ import print_function, unicode_literals

import glob
import os
import re
import time
from collections import defaultdict
from itertools import chain

from bgmi.config import MAX_PAGE, IS_PYTHON3, GLOBAL_FILTER, ENABLE_GLOBAL_FILTER
from bgmi.lib.models import Filter, Subtitle, STATUS_FOLLOWED, STATUS_UPDATED, Bangumi, STATUS_UPDATING
from bgmi.utils import (parse_episode, print_warning, print_info,
                        test_connection, download_cover, convert_cover_url_to_path)

if IS_PYTHON3:
    _unicode = str
else:
    _unicode = unicode


class BaseWebsite(object):
    cover_url = ''
    parse_episode = staticmethod(parse_episode)

    @staticmethod
    def save_data(data):
        """
        save bangumi dict to database

        :type data: dict
        """
        b, obj_created = Bangumi.get_or_create(name=data['name'], defaults=data)
        if not obj_created:
            b.status = STATUS_UPDATING
            b.subtitle_group = Bangumi(**data).subtitle_group
            b.cover = data['cover']
            # if not b.cover.startswith(self.cover_url):
            #     b.cover = self.cover_url + data['cover']
            b.save()

    def fetch(self, save=False, group_by_weekday=True):
        bangumi_result, subtitle_group_result = self.fetch_bangumi_calendar_and_subtitle_group()
        if subtitle_group_result:
            for subtitle_group in subtitle_group_result:
                s, if_created = Subtitle.get_or_create(id=_unicode(subtitle_group['id']),
                                                       name=_unicode(subtitle_group['name']))
                if if_created:
                    s.save()
        if not bangumi_result:
            print('no result return None')
            return []

        for bangumi in bangumi_result:
            bangumi['cover'] = self.cover_url + bangumi['cover']

        if save:
            for bangumi in bangumi_result:
                self.save_data(bangumi)
        if group_by_weekday:
            result_group_by_weekday = defaultdict(list)
            for bangumi in bangumi_result:
                result_group_by_weekday[bangumi['update_time'].lower()].append(bangumi)
            bangumi_result = result_group_by_weekday
        return bangumi_result

    @staticmethod
    def followed_bangumi():
        """

        :return: list of bangumi followed
        :rtype: list[dict]
        """
        weekly_list_followed = Bangumi.get_updating_bangumi(status=STATUS_FOLLOWED)
        weekly_list_updated = Bangumi.get_updating_bangumi(status=STATUS_UPDATED)
        weekly_list = defaultdict(list)
        for k, v in chain(weekly_list_followed.items(), weekly_list_updated.items()):
            weekly_list[k].extend(v)
        for bangumi_list in weekly_list.values():
            for bangumi in bangumi_list:
                bangumi['subtitle_group'] = [{'name': x['name'],
                                              'id': x['id']} for x in
                                             Subtitle.get_subtitle_by_id(bangumi['subtitle_group'].split(', '))]
        return weekly_list

    def bangumi_calendar(self, force_update=False, save=True, cover=None):
        """

        :param force_update:
        :type force_update: bool

        :param save: set true to enable save bangumi data to database
        :type save: bool

        :param cover: list of cover url (of scripts) want to download
        :type cover: list[str]
        """
        if force_update and not test_connection():
            force_update = False
            print_warning('Network is unreachable')

        if force_update:
            print_info('Fetching bangumi info ...')
            Bangumi.delete_all()
            weekly_list = self.fetch(save=save)
        else:
            weekly_list = Bangumi.get_updating_bangumi()
        if not weekly_list:
            print_warning('Warning: no bangumi schedule, fetching ...')
            weekly_list = self.fetch(save=save)

        if cover is not None:
            # download cover to local
            cover_to_be_download = cover
            for daily_bangumi in weekly_list.values():
                for bangumi in daily_bangumi:
                    _, file_path = convert_cover_url_to_path(bangumi['cover'])

                    if not glob.glob(file_path):
                        cover_to_be_download.append(bangumi['cover'])

            if cover_to_be_download:
                print_info('Updating cover ...')
                download_cover(cover_to_be_download)

        return weekly_list

    def get_maximum_episode(self, bangumi, subtitle=True, ignore_old_row=True, max_page=MAX_PAGE):
        """

        :type max_page: str
        :param max_page: 
        :type bangumi: object
        :type ignore_old_row: bool
        :param ignore_old_row: 
        :type bangumi: Bangumi
        :param subtitle: 
        :type subtitle: bool
        """
        try:
            followed_filter_obj = Filter.get(bangumi_name=bangumi.name)
        except Filter.DoesNotExist:
            followed_filter_obj = Filter.create(bangumi_name=bangumi.name)

        if followed_filter_obj and subtitle:
            subtitle_group = followed_filter_obj.subtitle
        else:
            subtitle_group = None

        if followed_filter_obj and subtitle:
            include = followed_filter_obj.include
        else:
            include = None

        if followed_filter_obj and subtitle:
            exclude = followed_filter_obj.exclude
        else:
            exclude = None

        if followed_filter_obj and subtitle:
            regex = followed_filter_obj.regex
        else:
            regex = None

        data = [i for i in self.fetch_episode(_id=bangumi.keyword, name=bangumi.name,
                                              subtitle_group=subtitle_group,
                                              include=include,
                                              exclude=exclude,
                                              regex=regex,
                                              max_page=int(max_page))
                if i['episode'] is not None]

        if ignore_old_row:
            data = [row for row in data if row['time'] > int(time.time()) - 3600 * 24 * 30 * 3]  # three month

        if data:
            bangumi = max(data, key=lambda _i: _i['episode'])
            return bangumi, data
        else:
            return {'episode': 0}, []

    def fetch_episode(self, _id, name='',
                      subtitle_group=None,
                      include=None,
                      exclude=None,
                      regex=None,
                      max_page=int(MAX_PAGE)):
        """
        :type _id: str
        :param _id:
        :type name: str
        :type subtitle_group: str
        :type include: str
        :type exclude: str
        :type regex: str
        :type max_page: int
        """
        result = []

        max_page = int(max_page)

        if subtitle_group and subtitle_group.split(', '):
            condition = subtitle_group.split(', ')
            response_data = self.fetch_episode_of_bangumi(bangumi_id=_id, subtitle_list=condition)
        else:
            response_data = self.fetch_episode_of_bangumi(bangumi_id=_id, max_page=max_page)

        for info in response_data:
            if '合集' not in info['title']:
                info['name'] = name
                result.append(info)

        if include:
            include_list = list(map(lambda s: s.strip(), include.split(',')))
            result = list(filter(lambda s: True if all(map(lambda t: t in s['title'],
                                                           include_list)) else False, result))

        if exclude:
            exclude_list = list(map(lambda s: s.strip(), exclude.split(',')))
            result = list(filter(lambda s: True if all(map(lambda t: t not in s['title'],
                                                           exclude_list)) else False, result))

        result = self.filter_keyword(data=result, regex=regex)
        return result

    @staticmethod
    def remove_duplicated_bangumi(result):
        """

        :type result: list[dict]
        """
        ret = []
        episodes = list({i['episode'] for i in result})
        for i in result:
            if i['episode'] in episodes:
                ret.append(i)
                del episodes[episodes.index(i['episode'])]

        return ret

    @staticmethod
    def filter_keyword(data, regex=None):
        """

        :type regex: str
        :param data: list of bangumi dict
        :type data: list[dict]
        """
        if regex:
            try:
                match = re.compile(regex)
                data = [s for s in data if match.findall(s['title'])]
            except re.error as e:
                if os.getenv('DEBUG'):  # pragma: no cover
                    import traceback
                    traceback.print_exc()
                    raise e
                return data

        if not ENABLE_GLOBAL_FILTER == '0':
            data = list(filter(lambda s: all(map(lambda t: t.strip().lower() not in s['title'].lower(),
                                                 GLOBAL_FILTER.split(','))), data))

        return data

    def search_by_keyword(self, keyword, count):
        """
        return a list of dict with at least 4 key: download, name, title, episode
        example:
        ```
            [
                {
                    'name':"路人女主的养成方法",
                    'download': 'magnet:?xt=urn:btih:what ever',
                    'title': "[澄空学园] 路人女主的养成方法 第12话 MP4 720p  完",
                    'episode': 12
                },
            ]

        :param keyword: search key word
        :type keyword: str
        :param count: how many page to fetch from website
        :type count: int

        :return: list of episode search result
        :rtype: list[dict]
        """
        raise NotImplementedError

    def fetch_bangumi_calendar_and_subtitle_group(self):
        """
        return a list of all bangumi and a list of all subtitle group

        list of bangumi dict:
        update time should be one of ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']
        example:
        ```
            [
                {
                    "status": 0,
                    "subtitle_group": [
                        "123",
                        "456"
                    ],
                    "name": "名侦探柯南",
                    "keyword": "1234", #bangumi id
                    "update_time": "Sat",
                    "cover": "data/images/cover1.jpg"
                },
            ]
        ```

        list of subtitle group dict:
        example:
        ```
            [
                {
                    'id': '233',
                    'name': 'bgmi字幕组'
                }
            ]
        ```


        :return: list of bangumi, list of subtitile group
        :rtype: (list[dict], list[dict])
        """
        raise NotImplementedError

    def fetch_episode_of_bangumi(self, bangumi_id, subtitle_list=None, max_page=MAX_PAGE):
        """
        get all episode by bangumi id
        example
        ```
            [
                {
                    "download": "magnet:?xt=urn:btih:e43b3b6b53dd9fd6af1199e112d3c7ff15cab82c",
                    "subtitle_group": "58a9c1c9f5dc363606ab42ec",
                    "title": "【喵萌奶茶屋】★七月新番★[来自深渊/Made in Abyss][07][GB][720P]",
                    "episode": 0,
                    "time": 1503301292
                },
            ]
        ```

        :param bangumi_id: bangumi_id
        :param subtitle_list: list of subtitle group
        :type subtitle_list: list
        :param max_page: how many page you want to crawl if there is no subtitle list
        :type max_page: int
        :return: list of bangumi
        :rtype: list[dict]
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import json
import logging
import os
import random
import string

from tornado.testing import AsyncHTTPTestCase

from bgmi.front.server import make_app
from bgmi.lib.constants import unicode_
from bgmi.config import SAVE_PATH

logging.basicConfig(level=logging.DEBUG)


def random_word(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))


logger = logging.getLogger()
logger.setLevel(logging.ERROR)
api_list = [
    {
        'action': 'update',
        'method': 'post',
        'params': '{}',
    }, {
        'action': 'status',
        'method': 'post',
        'params': json.dumps({
            'name': os.environ.get('BANGUMI_2'),
            'status': 1,
        }),
    }
]


# DB.recreate_source_relatively_table()


class ApiTestCase(AsyncHTTPTestCase):
    # class ApiTestCase(unittest.TestCase)
    headers = {'BGmi-Token': '233', 'Content-Type': 'application/json'}
    bangumi_1 = unicode_(os.environ.get('BANGUMI_1'))
    bangumi_2 = unicode_(os.environ.get('BANGUMI_2'))
    bangumi_3 = unicode_(os.environ.get('BANGUMI_3'))

    def get_app(self):
        self.app = make_app(debug=False)
        return self.app

    def test_404(self):
        pass
        # r = self.fetch('/api/url_does_not_exist')
        # self.assertEqual(r.code, 404)
        # res = self.parse_response(r)
        # self.assertEqual(res['status'], 'error')

    # def test_405(self):
    #     r = self.fetch('/url_does_not_exist')
    #     self.assertEqual(r.code, 405)
    #     res = self.parse_response(r)
    #     self.assertEqual(res['status'], 'error')

    def test_a_auth(self):
        r = self.fetch('/api/auth', method='POST', body=json.dumps({'token': '233'}))
        self.assertEqual(r.code, 200)
        res = self.parse_response(r)
        self.assertEqual(res['status'], 'success')

        r = self.fetch('/api/auth', method='POST', body=json.dumps({'token': '3'}))
        self.assertEqual(r.code, 400)
        res = self.parse_response(r)
        self.assertEqual(res['status'], 'error')

    def test_a_cal(self):
        r = self.fetch('/api/cal', method='GET')
        res = self.parse_response(r)
        self.assertIsInstance(res['data'], dict)

    def test_b_add(self):
        r = self.fetch('/api/add', method='POST',
                       headers=self.headers,
                       body=json.dumps({
                           'name': self.bangumi_1,
                       }))
        self.assertEqual(r.code, 200)

        r = self.fetch('/api/add', method='POST',
                       headers=self.headers,
                       body=json.dumps({
                           'name': self.bangumi_1,
                       }))
        self.assertEqual(r.code, 200)
        r = self.parse_response(r)
        self.assertEqual(r['status'], 'warning')

        r = self.fetch('/api/add', method='POST',
                       headers=self.headers,
                       body=json.dumps({
                           'name': self.bangumi_2,
                       }))
        self.assertEqual(r.code, 200)

    def test_c_delete(self):
        r = self.fetch('/api/add', method='POST',
                       headers=self.headers,
                       body=json.dumps({
                           'name': self.bangumi_2,
                       }))
        self.assertEqual(r.code, 200)
        r = self.parse_response(r)
        self.assertEqual(r['status'], 'warning')

        r = self.fetch('/api/add', method='POST',
                       headers=self.headers,
                       body=json.dumps({
                           'name': self.bangumi_2,
                       }))
        self.assertEqual(r.code, 200)
        r = self.parse_response(r)
        self.assertEqual(r['status'], 'warning')

        r = self.fetch('/api/add', method='POST',
                       headers=self.headers,
                       body=json.dumps({
                           'name': self.bangumi_2,
                       }))
        self.assertEqual(r.code, 200)
        r = self.parse_response(r)
        self.assertEqual(r['status'], 'warning')

    def test_e_mark(self):
        episode = random.randint(0, 10)
        self.fetch('/api/mark', method='POST', headers=self.headers,
                   body=json.dumps({
                       "name": self.bangumi_1,
                       "episode": episode
                   }))
        r = self.fetch('/api/index', method='GET')
        self.assertEqual(r.code, 200)
        res = self.parse_response(r)
        bg_dict = {}
        for item in res['data']:
            if item['bangumi_name'] == self.bangumi_1:
                bg_dict = item
                break
        self.assertEqual(bg_dict['bangumi_name'], self.bangumi_1)
        self.assertEqual(bg_dict['episode'], episode)

    def test_d_filter(self):
        include = random_word(5)
        exclude = random_word(5)
        regex = random_word(5)

        r = self.fetch('/api/filter', method='POST', body=json.dumps({
            'name': self.bangumi_1,
        }), headers=self.headers)

        self.assertEqual(r.code, 200)
        res = self.parse_response(r)
        self.assertEqual(res['status'], 'success')

        if len(res['data']['subtitle_group']) >= 2:
            subtitle_group = res['data']['subtitle_group'][:1]
        elif len(res['data']['subtitle_group']) == 1:
            subtitle_group = res['data']['subtitle_group'][:0]
        else:
            subtitle_group = []
        subtitle = ','.join(subtitle_group)

        r = self.fetch('/api/filter', method='POST', body=json.dumps({
            'name': self.bangumi_1,
            'include': include,
            'regex': regex,
            'exclude': exclude,
            'subtitle': subtitle,
        }), headers=self.headers)
        r = self.fetch('/api/filter', method='POST', body=json.dumps({
            'name': self.bangumi_1,
        }), headers=self.headers)

        res = self.parse_response(r)

        self.assertEqual(r.code, 200)
        self.assertEqual(res['status'], 'success')
        self.assertEqual(res['data']['name'], self.bangumi_1)
        self.assertEqual(res['data']['include'], include)
        self.assertEqual(res['data']['regex'], regex)
        self.assertEqual(res['data']['exclude'], exclude)

        r = self.fetch('/api/filter', method='POST', body=json.dumps({
            'name': self.bangumi_3,
            'regex': '.*',
            'subtitle': '',
        }), headers=self.headers)
        self.assertEqual(r.code, 400)
        self.assertEqual(self.parse_response(r)['status'], 'error')
        for item in subtitle_group:
            self.assertIn(item, res['data']['followed'])
        for item in res['data']['followed']:
            self.assertIn(item, subtitle_group)

    def test_e_index(self):
        save_dir = os.path.join(SAVE_PATH)
        episode1_dir = os.path.join(save_dir, self.bangumi_1, '1', 'episode1')
        if not os.path.exists(episode1_dir):
            os.makedirs(episode1_dir)
        open(os.path.join(episode1_dir, '1.mp4'), 'a').close()

        episode2_dir = os.path.join(save_dir, self.bangumi_1, '2')
        if not os.path.exists(episode2_dir):
            os.makedirs(episode2_dir)
        open(os.path.join(episode2_dir, '2.mkv'), 'a').close()

        response = self.fetch('/api/index', method='GET')
        self.assertEqual(response.code, 200)
        r = self.parse_response(response)
        episode_list = [x for x in r['data'] if x["bangumi_name"] == self.bangumi_1]
        bangumi_dict = next(iter(episode_list or []), {})

        self.assertIn('1', bangumi_dict['player'].keys())
        self.assertEqual(bangumi_dict['player']['1']['path'], '/{}/1/episode1/1.mp4'.format(self.bangumi_1))
        self.assertIn('2', bangumi_dict['player'].keys())
        self.assertEqual(bangumi_dict['player']['2']['path'], '/{}/2/2.mkv'.format(self.bangumi_1))

    def test_f_update(self):
        return
        r = self.fetch('/api/update', method='POST', headers=self.headers,
                       body=json.dumps({
                           "name": ''
                       }))
        res = self.parse_response(r)
        self.assertEqual(res['status'], 'info')

    def test_resource_ics(self):
        r = self.fetch('/resource/feed.xml')
        self.assertEqual(r.code, 200)

    def test_resource_feed(self):
        r = self.fetch('/resource/calendar.ics')
        self.assertEqual(r.code, 200)

    def test_no_auth(self):
        r = self.fetch('/api/add', method='POST', body=json.dumps({'name': self.bangumi_1}))
        self.assertEqual(r.code, 401)

    @staticmethod
    def parse_response(response):
        r = json.loads(response.body.decode('utf-8'))
        return r



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import os
import unittest

from bgmi.config import unicode_
from bgmi.lib.fetch import website


class ControllersTest(unittest.TestCase):
    def setUp(self):
        self.bangumi_name_1 = unicode_(os.environ.get('BANGUMI_1'))
        self.bangumi_name_2 = unicode_(os.environ.get('BANGUMI_2'))
        self.w = website
        pass

    def test_info(self):
        bs, gs = self.w.fetch_bangumi_calendar_and_subtitle_group()
        b = {}
        for bangumi in bs:
            self.assertIn("status", bangumi)
            self.assertIn("subtitle_group", bangumi)
            self.assertIn("name", bangumi)
            self.assertIn("keyword", bangumi)
            self.assertIn("update_time", bangumi)
            self.assertIn("cover", bangumi)
            if bangumi['name'] == self.bangumi_name_1:
                b = bangumi

        for subtitle_group in gs:
            self.assertIn('id', subtitle_group)
            self.assertIn('name', subtitle_group)
        self.assertTrue(bool(b))
        es = self.w.fetch_episode_of_bangumi(b['keyword'])
        for episode in es:
            self.assertIn('download', episode)
            self.assertIn('subtitle_group', episode)
            self.assertIn('title', episode)
            self.assertIn('episode', episode)
            self.assertIn('time', episode)

    def test_search(self):
        r = self.w.search_by_keyword(self.bangumi_name_1)
        for b in r:
            self.assertIn('name', b)
            self.assertIn('download', b)
            self.assertIn('title', b)
            self.assertIn('episode', b)

    # @staticmethod
    # def setUpClass():
    #     setup()
    #     recreate_source_relatively_table()



# coding=utf-8
from __future__ import print_function, unicode_literals

import unittest
from bgmi.lib.controllers import config


class ConfigTest(unittest.TestCase):
    def test_print_config(self):
        r = config()
        self.assertEqual(r['status'], 'info')

    def test_print_single_config(self):
        r = config('DANMAKU_API_URL', '233')

        r = config('DANMAKU_API_URL')
        self.assertEqual(r['status'], 'info')
        self.assertTrue(r['message'].startswith('DANMAKU_API_URL'))
        self.assertTrue(r['message'].endswith('233'))

    def test_readonly(self):
        r = config('DB_PATH', '/tmp/233')
        self.assertEqual(r['status'], 'error')

    def test_writable(self):
        r = config('DANMAKU_API_URL', '233')
        self.assertEqual(r['status'], 'success')
        from bgmi.config import DANMAKU_API_URL
        self.assertEqual(DANMAKU_API_URL, '233')

    def test_source(self):
        r = config('DATA_SOURCE', '233')
        self.assertEqual(r['status'], 'error')

    def test_wrong_config_name(self):
        r = config('WRONG_CONFIG_NAME', '233')
        self.assertEqual(r['status'], 'error')

    # def value_data(self):
    def test_wrong_DOWNLOAD_DELEGATE(self):
        r = config('DOWNLOAD_DELEGATE', 'WRONG_METHOD')
        self.assertEqual(r['status'], 'error')

    def test_DOWNLOAD_DELEGATE(self):
        r = config('DOWNLOAD_DELEGATE', 'aria2-rpc')
        r = config('DOWNLOAD_DELEGATE', 'rr!')
        r = config('WGET_PATH', 'some_place')
        # self.assertEqual(r['status'], 'error')

    @staticmethod
    def setUpClass():
        from bgmi.config import CONFIG_FILE_PATH
        import os
        os.remove(CONFIG_FILE_PATH)
        from bgmi.config import write_default_config
        write_default_config()

    @staticmethod
    def tearDownClass():
        from bgmi.config import CONFIG_FILE_PATH
        import os
        os.remove(CONFIG_FILE_PATH)
        from bgmi.config import write_default_config
        write_default_config()



# coding=utf-8
from __future__ import print_function, unicode_literals

import time

from bgmi.config import write_config, MAX_PAGE
from bgmi.lib.constants import SUPPORT_WEBSITE
from bgmi.lib.download import download_prepare
from bgmi.lib.fetch import website
from bgmi.lib.models import (Filter, Subtitle, Download, recreate_source_relatively_table,
                             STATUS_FOLLOWED, STATUS_UPDATED, STATUS_NOT_DOWNLOAD, FOLLOWED_STATUS, Followed, Bangumi,
                             DoesNotExist, model_to_dict)
from bgmi.lib.models import (STATUS_DELETED)
from bgmi.script import ScriptRunner
from bgmi.utils import print_info, normalize_path, print_warning, print_success, print_error, GREEN, COLOR_END


def add(name, episode=None):
    """
    ret.name :str
    """
    # action add
    # add bangumi by a list of bangumi name
    # result = {}
    if not Bangumi.get_updating_bangumi():
        website.fetch(save=True, group_by_weekday=False)

    try:
        bangumi_obj = Bangumi.get(name=name)
    except Bangumi.DoesNotExist:
        return {'status': 'error',
                'message': '{0} not found, please check the name'.format(name)}

    followed_obj, this_obj_created = Followed.get_or_create(bangumi_name=bangumi_obj.name,
                                                            defaults={'status': STATUS_FOLLOWED})
    if not this_obj_created:
        if followed_obj.status == STATUS_FOLLOWED:
            return {'status': 'warning', 'message': '{0} already followed'.format(bangumi_obj.name)}
        else:
            followed_obj.status = STATUS_FOLLOWED
            followed_obj.save()

    Filter.get_or_create(bangumi_name=name)

    bangumi_data, _ = website.get_maximum_episode(bangumi_obj, subtitle=False, max_page=1)
    followed_obj.episode = bangumi_data['episode'] if episode is None else episode
    followed_obj.save()
    result = {'status': 'success', 'message': '{0} has been followed'.format(bangumi_obj.name)}

    return result


def filter_(name, subtitle=None, include=None, exclude=None, regex=None):
    result = {'status': 'success', 'message': ''}
    try:
        bangumi_obj = Bangumi.get(name=name)
    except Bangumi.DoesNotExist:
        result['status'] = 'error'
        result['message'] = 'Bangumi {0} does not exist.'.format(name)
        return result

    try:
        Followed.get(bangumi_name=bangumi_obj.name)
    except Followed.DoesNotExist as exc:
        result['status'] = 'error'
        result['message'] = 'Bangumi {name} has not subscribed, try \'bgmi add "{name}"\'.' \
            .format(name=bangumi_obj.name)
        return result

    followed_filter_obj, is_this_obj_created = Filter.get_or_create(bangumi_name=name)

    if is_this_obj_created:
        followed_filter_obj.save()

    if subtitle is not None:
        subtitle = [s.strip() for s in subtitle.split(',')]
        subtitle = [s['id'] for s in Subtitle.get_subtitle_by_name(subtitle)]
        subtitle_list = [s.split('.')[0] for s in bangumi_obj.subtitle_group.split(', ') if '.' in s]
        subtitle_list.extend(bangumi_obj.subtitle_group.split(', '))
        subtitle = filter(lambda s: s in subtitle_list, subtitle)
        subtitle = ', '.join(subtitle)
        followed_filter_obj.subtitle = subtitle

    if include is not None:
        followed_filter_obj.include = include

    if exclude is not None:
        followed_filter_obj.exclude = exclude

    if regex is not None:
        followed_filter_obj.regex = regex

    followed_filter_obj.save()
    subtitle_list = list(map(lambda s: s['name'],
                             Subtitle.get_subtitle_by_id(bangumi_obj.subtitle_group.split(', '))))

    result['data'] = {
        'name': name,
        'subtitle_group': subtitle_list,
        'followed': list(map(lambda s: s['name'], Subtitle.get_subtitle_by_id(followed_filter_obj.subtitle.split(', ')))
                         if followed_filter_obj.subtitle else []),
        'include': followed_filter_obj.include,
        'exclude': followed_filter_obj.exclude,
        'regex': followed_filter_obj.regex,
    }
    return result


def delete(name='', clear_all=False, batch=False):
    """
    :param name:
    :type name: unicode
    :param clear_all:
    :type clear_all: bool
    :param batch:
    :type batch: bool
    :return:
    """
    # action delete
    # just delete subscribed bangumi or clear all the subscribed bangumi
    result = {}
    if clear_all:
        if Followed.delete_followed(batch=batch):
            result['status'] = "warning"
            result['message'] = 'all subscriptions have been deleted'
        else:
            print_error('user canceled')
    elif name:
        try:
            followed = Followed.get(bangumi_name=name)
            followed.status = STATUS_DELETED
            followed.save()
            result['status'] = 'warning'
            result['message'] = 'Bangumi {} has been deleted'.format(name)
        except Followed.DoesNotExist:
            result['status'] = 'error'
            result['message'] = 'Bangumi %s does not exist' % name
    else:
        result['status'] = 'warning'
        result['message'] = 'Nothing has been done.'
    return result


def cal(force_update=False, save=False):
    weekly_list = website.bangumi_calendar(force_update=force_update, save=save)
    runner = ScriptRunner()
    patch_list = runner.get_models_dict()
    for i in patch_list:
        weekly_list[i['update_time'].lower()].append(i)

    # for web api
    r = weekly_list
    for day, value in weekly_list.items():
        for index, bangumi in enumerate(value):
            bangumi['cover'] = normalize_path(bangumi['cover'])
            if isinstance(bangumi['subtitle_group'], list):
                subtitle_group = list(map(lambda x: {'name': x['name'], 'id': x['id']},
                                          Subtitle.get_subtitle_by_id(
                                              bangumi['subtitle_group'])))
            else:
                subtitle_group = list(map(lambda x: {'name': x['name'], 'id': x['id']},
                                          Subtitle.get_subtitle_by_id(
                                              bangumi['subtitle_group'].split(', ' ''))))

            r[day][index]['subtitle_group'] = subtitle_group
    return r


def download(name, title, episode, download_url):
    my_dict = {
        'name': name,
        'title': title,
        'episode': episode,
        'download': download_url,
    }
    download_prepare(my_dict)


def mark(name, episode):
    """

    :param name: name of the bangumi you want to mark
    :type name: str
    :param episode: bangumi episode you want to mark
    :type episode: int
    :return: result
    :rtype: dict[status: str,message: str]
    """
    result = {}
    try:
        followed_obj = Followed.get(bangumi_name=name)
    except Followed.DoesNotExist:
        runner = ScriptRunner()
        followed_obj = runner.get_model(name)
        if not followed_obj:
            result['status'] = 'error'
            result['message'] = 'Subscribe or Script <{}> does not exist.'.format(name)
            return result

    if episode is not None:
        followed_obj.episode = episode
        followed_obj.save()
        result['status'] = 'success'
        result['message'] = '{} has been mark as episode: {}'.format(name, episode)
    else:  # episode is None
        result['status'] = 'info'
        result['message'] = '{}, episode: {}'.format(name, followed_obj.episode)
    return result


def search(keyword, count=MAX_PAGE, regex=None, dupe=True):
    try:
        count = int(count)
    except (TypeError, ValueError):
        count = 3

    data = website.search_by_keyword(keyword, count=count)
    if not dupe:
        data = website.remove_duplicated_bangumi(data)
    else:
        data = website.filter_keyword(data, regex=regex)

    return data


def source(data_source):
    result = {}
    if data_source in list(map(lambda x: x['id'], SUPPORT_WEBSITE)):
        recreate_source_relatively_table()
        write_config('DATA_SOURCE', data_source)
        print_success('data source switch succeeds')
        result['status'] = 'success'
        result['message'] = 'you have successfully change your data source to {}'.format(data_source)
    else:
        result['status'] = 'error'
        result['message'] = 'please check input.nata source should be {} or {}'.format(
            *[x['id'] for x in SUPPORT_WEBSITE])
    return result


def config(name=None, value=None):
    if name == 'DATA_SOURCE':
        error_message = "you can't change data source in this way. please use `bgmi source ${data source}` in cli"
        result = {'status': 'error',
                  'message': error_message,
                  'data': write_config()['data']}
        return result
    r = write_config(name, value)
    if name == 'ADMIN_TOKEN':
        r['message'] = 'you need to restart your bgmi_http to make new token work'
    return r


def update(name, download=None, not_ignore=False):
    result = {'status': 'info', 'message': '', 'data': {'updated': [], 'downloaded': []}}

    ignore = not bool(not_ignore)
    print_info('marking bangumi status ...')
    now = int(time.time())
    for i in Followed.get_all_followed():
        if i['updated_time'] and int(i['updated_time'] + 86400) < now:
            followed_obj = Followed.get(bangumi_name=i['bangumi_name'])
            followed_obj.status = STATUS_FOLLOWED
            followed_obj.save()

    for script in ScriptRunner().scripts:
        obj = script.Model().obj
        if obj.updated_time and int(obj.updated_time + 86400) < now:
            obj.status = STATUS_FOLLOWED
            obj.save()

    print_info('updating bangumi data ...')
    website.fetch(save=True, group_by_weekday=False)
    print_info('updating subscriptions ...')
    download_queue = []

    if download:
        if not name:
            print_warning('No specified bangumi, ignore `--download` option')
        if len(name) > 1:
            print_warning(
                'Multiple specified bangumi, ignore `--download` option')

    if not name:
        updated_bangumi_obj = Followed.get_all_followed()
    else:
        updated_bangumi_obj = []
        for i in name:
            try:
                f = Followed.get(bangumi_name=i)
                f = model_to_dict(f)
                updated_bangumi_obj.append(f)
            except DoesNotExist:
                pass

    runner = ScriptRunner()
    script_download_queue = runner.run()

    for subscribe in updated_bangumi_obj:
        print_info('fetching %s ...' % subscribe['bangumi_name'])
        try:
            bangumi_obj = Bangumi.get(name=subscribe['bangumi_name'])
        except Bangumi.DoesNotExist:
            print_error('Bangumi<{0}> does not exists.'.format(subscribe['bangumi_name']),
                        exit_=False)
            continue
        try:
            followed_obj = Followed.get(bangumi_name=subscribe['bangumi_name'])
        except Followed.DoesNotExist:
            print_error('Bangumi<{0}> is not followed.'.format(subscribe['bangumi_name']),
                        exit_=False)
            continue

        episode, all_episode_data = website.get_maximum_episode(
            bangumi=bangumi_obj, ignore_old_row=ignore, max_page=1)

        if (episode.get('episode') > subscribe['episode']) or (len(name) == 1 and download):
            if len(name) == 1 and download:
                episode_range = download
            else:
                episode_range = range(
                    subscribe['episode'] + 1, episode.get('episode', 0) + 1)
                print_success('%s updated, episode: %d' %
                              (subscribe['bangumi_name'], episode['episode']))
                followed_obj.episode = episode['episode']
                followed_obj.status = STATUS_UPDATED
                followed_obj.updated_time = int(time.time())
                followed_obj.save()
                result['data']['updated'].append({'bangumi': subscribe['bangumi_name'],
                                                  'episode': episode['episode']})

            for i in episode_range:
                for epi in all_episode_data:
                    if epi['episode'] == i:
                        download_queue.append(epi)
                        break

    if download is not None:
        result['data']['downloaded'] = download_queue
        download_prepare(download_queue)
        download_prepare(script_download_queue)
        print_info('Re-downloading ...')
        download_prepare(Download.get_all_downloads(
            status=STATUS_NOT_DOWNLOAD))

    return result


def status_(name, status=STATUS_DELETED):
    result = {'status': 'success', 'message': ''}

    if not status in FOLLOWED_STATUS or not status:
        result['status'] = 'error'
        result['message'] = 'Invalid status: {0}'.format(status)
        return result

    status = int(status)
    try:
        followed_obj = Followed.get(bangumi_name=name)
    except Followed.DoesNotExist:
        result['status'] = 'error'
        result['message'] = 'Followed<{0}> does not exists'.format(name)
        return result

    followed_obj.status = status
    followed_obj.save()
    result['message'] = 'Followed<{0}> has been marked as status {1}'.format(name, status)
    return result


def list_():
    result = {}
    weekday_order = Bangumi.week
    followed_bangumi = website.followed_bangumi()

    script_bangumi = ScriptRunner().get_models_dict()

    if not followed_bangumi and not script_bangumi:
        result['status'] = 'warning'
        result['message'] = 'you have not subscribed any bangumi'
        return result

    for i in script_bangumi:
        i['subtitle_group'] = [{'name': '<BGmi Script>'}]
        followed_bangumi[i['update_time'].lower()].append(i)

    result['status'] = 'info'
    result['message'] = ''
    for index, weekday in enumerate(weekday_order):
        if followed_bangumi[weekday.lower()]:
            result['message'] += '%s%s. %s' % (GREEN, weekday, COLOR_END)
            for i, bangumi in enumerate(followed_bangumi[weekday.lower()]):
                if bangumi['status'] in (STATUS_UPDATED, STATUS_FOLLOWED) and 'episode' in bangumi:
                    bangumi['name'] = '%s(%d)' % (
                        bangumi['name'], bangumi['episode'])
                if i > 0:
                    result['message'] += ' ' * 5
                f = map(lambda x: x['name'], bangumi['subtitle_group'])
                result['message'] += '%s: %s\n' % (bangumi['name'], ', '.join(f) if f else '<None>')

    return result



# coding: utf-8
import functools
import traceback
from multiprocessing.pool import ThreadPool

from tornado.web import asynchronous, HTTPError

from bgmi.config import ADMIN_TOKEN
from bgmi.lib.constants import (ACTION_ADD, ACTION_DELETE, ACTION_CAL, ACTION_SEARCH, ACTION_CONFIG, ACTION_DOWNLOAD,
                                ACTION_MARK, ACTION_FILTER)
from bgmi.lib.controllers import add, delete, search, cal, config, update, mark, status_, filter_
from bgmi.lib.download import download_prepare
from bgmi.front.base import BaseHandler

ACTION_AUTH = 'auth'
ACTION_STATUS = 'status'


def auth_(token=''):
    return {'status': 'success' if token == ADMIN_TOKEN else 'error'}


API_MAP_POST = {
    ACTION_ADD: add,
    ACTION_DELETE: delete,
    ACTION_SEARCH: search,
    ACTION_CONFIG: config,
    ACTION_DOWNLOAD: download_prepare,
    ACTION_AUTH: auth_,
    ACTION_MARK: mark,
    ACTION_STATUS: status_,
    ACTION_FILTER: filter_,
}

API_MAP_GET = {
    ACTION_CAL: lambda: {'data': cal()},
    ACTION_CONFIG: lambda: config(None, None)
}

NO_AUTH_ACTION = (ACTION_SEARCH, ACTION_CAL, ACTION_AUTH)


def auth(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if kwargs.get('action', None) in NO_AUTH_ACTION:
            return f(*args, **kwargs)

        token = args[0].request.headers.get('bgmi-token')
        if token == ADMIN_TOKEN:
            return f(*args, **kwargs)
        else:
            # HTTPError will be except in `BaseHandler.write_error`
            raise HTTPError(401)

    return wrapper


class AdminApiHandler(BaseHandler):
    @auth
    def get(self, action, *args, **kwargs):
        try:
            result = API_MAP_GET.get(action)()
        except Exception:
            traceback.print_exc()
            raise HTTPError(400)
        self.finish(self.jsonify(**result))

    @auth
    def post(self, action, *args, **kwargs):
        data = self.get_json()

        try:
            result = API_MAP_POST.get(action)(**data)
            if result['status'] == 'error':
                raise HTTPError(400)
        except HTTPError:
            raise HTTPError(400)
        except Exception:
            traceback.print_exc()
            raise HTTPError(500)

        resp = self.jsonify(**result)
        self.finish(resp)


class UpdateHandler(BaseHandler):
    @auth
    @asynchronous
    def post(self):
        data = self.get_json()

        name = data.get('name', '')
        download = data.get('download', [])

        if not download:
            download = None

        pool = ThreadPool(processes=1)
        pool.apply_async(update, (name, download,), callback=self.resp)

    def resp(self, result):
        self.write(self.jsonify(**result))
        self.finish()



# encoding: utf-8
from __future__ import print_function, unicode_literals

import os

import tornado.httpserver
import tornado.ioloop
import tornado.options
import tornado.template
import tornado.web
from tornado.options import options, define

from bgmi.config import SAVE_PATH, FRONT_STATIC_PATH, TORNADO_SERVE_STATIC_FILES
from bgmi.front.admin import AdminApiHandler, UpdateHandler, API_MAP_POST, API_MAP_GET
from bgmi.front.index import BangumiListHandler, IndexHandler
from bgmi.front.resources import RssHandler, CalendarHandler, BangumiHandler

define('port', default=8888, help='listen on the port', type=int)
define('address', default='0.0.0.0', help='binding at given address', type=str)

API_ACTIONS = '%s|%s' % ('|'.join(API_MAP_GET.keys()), '|'.join(API_MAP_POST.keys()))

if os.environ.get('DEV'):  # pragma: no cover
    def prepare(self):
        self.set_header('Access-Control-Allow-Origin', 'http://localhost:8080')
        self.set_header("Access-Control-Allow-Methods", "GET,POST,PUT,DELETE,OPTIONS")
        self.set_header("Access-Control-Allow-Headers", "Content-Type, bgmi-token, X-Requested-With")


    tornado.web.RequestHandler.prepare = prepare


def make_app(**kwargs):
    settings = {
        'gzip': True,
        'debug': True,
    }
    settings.update(kwargs)
    handlers = [
        (r'^/api/(old|index)', BangumiListHandler),
        (r'^/resource/feed.xml$', RssHandler),
        (r'^/resource/calendar.ics$', CalendarHandler),
        (r'^/api/update', UpdateHandler),
        (r'^/api/(?P<action>%s)' % API_ACTIONS, AdminApiHandler),
    ]

    if TORNADO_SERVE_STATIC_FILES != '0':
        handlers.extend([
            (r'/bangumi/(.*)', tornado.web.StaticFileHandler, {'path': SAVE_PATH}),
            (r'^/(.*)$', tornado.web.StaticFileHandler, {'path': FRONT_STATIC_PATH,
                                                         'default_filename': 'index.html'})
        ])
    else:
        handlers.extend([
            (r'^/bangumi/?(.*)', BangumiHandler),
            (r'^/(.*)$', IndexHandler)
        ])

    return tornado.web.Application(handlers, **settings)


def main():
    tornado.options.parse_command_line()
    print('BGmi HTTP Server listening on %s:%d' % (options.address, options.port))
    http_server = tornado.httpserver.HTTPServer(make_app())
    http_server.listen(options.port, address=options.address)
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    main()



from __future__ import print_function, unicode_literals
import base64
from six import iteritems

from bgmi.downloader.base import BaseDownloadService
from bgmi.config import TRANSMISSION_RPC_PORT, TRANSMISSION_RPC_URL, IS_PYTHON3
from bgmi.utils import print_info, print_warning


if IS_PYTHON3:
    import urllib
    from urllib.parse import urlparse
else:
    import urllib2 as urllib
    from urlparse import urlparse


try:
    import transmissionrpc
    from transmissionrpc.utils import make_rpc_name, argument_value_convert

    class PatchClient(transmissionrpc.Client):
        def add_torrent(self, torrent, timeout=None, **kwargs):
            if torrent is None:
                raise ValueError('add_torrent requires data or a URI.')
            torrent_data = None
            parsed_uri = urlparse(torrent)
            if parsed_uri.scheme in ['ftp', 'ftps', 'http', 'https']:
                # there has been some problem with T's built in torrent fetcher,
                # use a python one instead
                opener = urllib.build_opener()
                opener.addheaders = [('User-Agent', 'BGmi/Torrent-Downloader')]
                torrent_file = opener.open(torrent)

                torrent_data = torrent_file.read()
                torrent_data = base64.b64encode(torrent_data).decode('utf-8')
            if parsed_uri.scheme in ['file']:
                filepath = torrent
                # uri decoded different on linux / windows ?
                if len(parsed_uri.path) > 0:
                    filepath = parsed_uri.path
                elif len(parsed_uri.netloc) > 0:
                    filepath = parsed_uri.netloc
                torrent_file = open(filepath, 'rb')
                torrent_data = torrent_file.read()
                torrent_data = base64.b64encode(torrent_data).decode('utf-8')
            if not torrent_data:
                if torrent.endswith('.torrent') or torrent.startswith('magnet:'):
                    torrent_data = None
                else:
                    might_be_base64 = False
                    try:
                        # check if this is base64 data
                        if IS_PYTHON3:
                            base64.b64decode(torrent.encode('utf-8'))
                        else:
                            base64.b64decode(torrent)
                        might_be_base64 = True
                    except Exception:
                        pass
                    if might_be_base64:
                        torrent_data = torrent
            args = {}
            if torrent_data:
                args = {'metainfo': torrent_data}
            else:
                args = {'filename': torrent}
            for key, value in iteritems(kwargs):
                argument = make_rpc_name(key)
                (arg, val) = argument_value_convert('torrent-add', argument, value, self.rpc_version)
                args[arg] = val
            return list(self._request('torrent-add', args, timeout=timeout).values())[0]
except ImportError:
    class PatchClient(object):
        pass


class TransmissionRPC(BaseDownloadService):

    def __init__(self, *args, **kwargs):
        self.check_delegate_bin_exist('')
        super(TransmissionRPC, self).__init__(*args, **kwargs)

    def download(self):
        try:
            import transmissionrpc
            tc = PatchClient(TRANSMISSION_RPC_URL, port=TRANSMISSION_RPC_PORT)
            tc.add_torrent(self.torrent, download_dir=self.save_path.encode('utf-8'))
            print_info('Add torrent into the download queue, the file will be saved at {0}'.format(self.save_path))
        except ImportError:
            pass

    def check_delegate_bin_exist(self, path):
        try:
            import transmissionrpc
        except ImportError:
            raise Exception('Please run `pip install transmissionrpc`')

    def check_download(self, name):
        pass

    @staticmethod
    def download_status(status=None):
        print_info('Print download status in database')
        BaseDownloadService.download_status(status=status)
        print('')
        print_info('Print download status in transmission-rpc')
        try:
            import transmissionrpc
            tc = transmissionrpc.Client(TRANSMISSION_RPC_URL, port=TRANSMISSION_RPC_PORT)
            for torrent in tc.get_torrents():
                print_info('  * {0}: {1}'.format(torrent.status, torrent), indicator=False)
        except ImportError:
            pass

    @staticmethod
    def install():
        print_warning('Please run `pip install transmissionrpc`')



# coding=utf-8
import os
from shutil import copy

from bgmi.config import (IS_WINDOWS, BGMI_PATH, DOWNLOAD_DELEGATE,
                         SAVE_PATH, FRONT_STATIC_PATH, TMP_PATH, SCRIPT_PATH, TOOLS_PATH)
from bgmi.lib.download import get_download_class
from bgmi.utils import print_success, print_warning, print_info, print_error


def install_crontab():
    print_info('Installing crontab job')
    if IS_WINDOWS:
        copy(os.path.join(os.path.dirname(__file__), 'others/cron.vbs'), BGMI_PATH)
        os.system('schtasks /Create /SC HOURLY /TN "bgmi updater" /TR "{}"  /IT /F'.format(
            os.path.join(BGMI_PATH, 'cron.vbs')))
    else:
        path = os.path.join(os.path.dirname(__file__), 'others/crontab.sh')
        os.system("bash '%s'" % path)


def create_dir():
    path_to_create = (BGMI_PATH, SAVE_PATH, TMP_PATH,
                      SCRIPT_PATH, TOOLS_PATH, FRONT_STATIC_PATH)

    if not os.environ.get('HOME', os.environ.get('USERPROFILE', '')):
        print_warning('$HOME not set, use \'/tmp/\'')

    # bgmi home dir
    try:
        for path in path_to_create:
            if not os.path.exists(path):
                os.makedirs(path)
                print_success('%s created successfully' % path)
    except OSError as e:
        print_error('Error: {0}'.format(str(e)))


def install():
    get_download_class(DOWNLOAD_DELEGATE, instance=False).install()



# coding=utf-8
from __future__ import print_function, unicode_literals

import os
import sys
import signal
import argparse

from bgmi.lib.cli import controllers
from bgmi.config import BGMI_PATH, IS_PYTHON3
from bgmi.lib.constants import actions_and_arguments, ACTION_COMPLETE
from bgmi.setup import create_dir, install_crontab
from bgmi.sql import init_db
from bgmi.lib.update import update_database
from bgmi.utils import print_warning, print_error, print_version, check_update, get_web_admin


# global Ctrl-C signal handler
def signal_handler(signal, frame):  # pragma: no cover
    print_error('User aborted, quit')


if not IS_PYTHON3 and sys.platform.startswith('win'):
    reload(sys)
    sys.setdefaultencoding('gbk')


signal.signal(signal.SIGINT, signal_handler)


# main function
def main():
    setup()
    c = argparse.ArgumentParser()

    c.add_argument('--version', help='Show the version of BGmi.', action='version', version=print_version())

    sub_parser = c.add_subparsers(help='BGmi actions', dest='action')

    for action in actions_and_arguments:
        tmp_sub_parser = sub_parser.add_parser(action['action'], help=action.get('help', ''))
        for sub_action in action.get('arguments', []):
            tmp_sub_parser.add_argument(sub_action['dest'], **sub_action['kwargs'])

    sub_parser.add_parser(ACTION_COMPLETE, help='Gen completion, `eval "$(bgmi complete)"` '
                                                'or `eval "$(bgmi complete|dos2unix)"`')
    # sub_parser_del.add_argument('command', nargs='+', )

    ret = c.parse_args()
    if ret.action == 'install':
        import bgmi.setup

        bgmi.setup.install()
        get_web_admin(method='install')

        raise SystemExit
    elif ret.action == 'upgrade':
        create_dir()
        update_database()
        check_update(mark=False)
    else:
        check_update()
        controllers(ret)


def setup():
    need_to_init = False
    if not os.path.exists(BGMI_PATH):
        need_to_init = True
        print_warning('BGMI_PATH %s does not exist, installing' % BGMI_PATH)

    create_dir()
    init_db()
    if need_to_init:
        install_crontab()


if __name__ == '__main__':
    setup()
    main()



# coding=utf-8
from __future__ import print_function, unicode_literals

import datetime
import os
import time

import requests

from bgmi.config import (LANG, MAX_PAGE, BANGUMI_MOE_URL)
from bgmi.lib.models import Bangumi
from bgmi.utils import (print_info, bug_report, print_error, print_warning)
from bgmi.website.base import BaseWebsite

# tag of bangumi on bangumi.moe
BANGUMI_TAG = '549ef207fe682f7549f1ea90'

__split = '/' if not BANGUMI_MOE_URL.endswith('/') else ''
FETCH_URL = '{0}{1}api/bangumi/current'.format(BANGUMI_MOE_URL, __split)
TEAM_URL = '{0}{1}api/team/working'.format(BANGUMI_MOE_URL, __split)
NAME_URL = '{0}{1}api/tag/fetch'.format(BANGUMI_MOE_URL, __split)
DETAIL_URL = '{0}{1}api/torrent/search'.format(BANGUMI_MOE_URL, __split)
SEARCH_URL = '{0}{1}api/v2/torrent/search'.format(BANGUMI_MOE_URL, __split)
TORRENT_URL = '{0}{1}download/torrent/'.format(BANGUMI_MOE_URL, __split)
COVER_URL = 'https://bangumi.moe/'


def get_response(url, method='GET', **kwargs):
    # kwargs['proxies'] = {'http': "http://localhost:1080"}
    if os.environ.get('DEV'):  # pragma: no cover
        url = url.replace('https://', 'http://localhost:8092/https/')
    if os.environ.get('DEBUG'):  # pragma: no cover
        print_info('Request URL: {0}'.format(url))
    try:
        r = requests.request(method.lower(), url, **kwargs)
        if os.environ.get('DEBUG'):  # pragma: no cover
            print(r.text)
        return r.json()
    except requests.ConnectionError:
        print_error('error: failed to establish a new connection')
    except ValueError:
        print_error('error: server returned data maybe not be json, please contact ricterzheng@gmail.com')


def process_name(data):
    lang = 'zh_cn' if LANG not in ('zh_cn', 'zh_tw', 'ja', 'en') else LANG
    return {i['_id']: i['locale'][lang] for i in data}


def process_subtitle(data):
    """get subtitle group name from links
    """
    result = {}
    for s in data:
        # pprint(data)
        # f = Subtitle(id=s['tag_id'], name=s['name'])
        # if not f.select():
        #     f.save()
        result[s['tag_id']] = s['name']
    return result


def parser_bangumi(data):
    """match weekly bangumi list from data
    """

    ids = list(map(lambda b: b['tag_id'], data))
    subtitle = get_response(TEAM_URL, 'POST', json={'tag_ids': ids})
    name = process_name(get_response(NAME_URL, 'POST', json={'_ids': ids}))

    weekly_list = []
    subtitle_group_list = []
    for bangumi_item in data:
        subtitle_of_bangumi = process_subtitle(subtitle.get(bangumi_item['tag_id'], []))
        item = {'status': 0,
                'subtitle_group': list(subtitle_of_bangumi.keys()),
                'name': name[bangumi_item['tag_id']],
                'keyword': bangumi_item['tag_id'],
                'update_time': Bangumi.week[bangumi_item['showOn'] - 1],
                'cover': bangumi_item['cover']}
        for key, value in subtitle_of_bangumi.items():
            subtitle_group_list.append({
                'id': key,
                'name': value
            })

        weekly_list.append(item)

    if not weekly_list:
        bug_report()
    return weekly_list, subtitle_group_list


class BangumiMoe(BaseWebsite):
    cover_url = COVER_URL

    def fetch_episode_of_bangumi(self, bangumi_id, subtitle_list=None, max_page=MAX_PAGE):
        response_data = []
        ret = []
        if subtitle_list:
            for subtitle_id in subtitle_list:
                data = {'tag_id': [bangumi_id, subtitle_id, BANGUMI_TAG]}
                response = get_response(DETAIL_URL, 'POST', json=data)
                response_data.extend(response['torrents'])
        else:
            for i in range(max_page):
                if max_page > 1:
                    print_info('Fetch page {0} ...'.format(i + 1))
                data = {'tag_id': [bangumi_id, BANGUMI_TAG], 'p': i + 1}
                response = get_response(DETAIL_URL, 'POST', json=data)
                if response:
                    response_data.extend(response['torrents'])

        for index, bangumi in enumerate(response_data):
            ret.append({
                # 'download': bangumi['magnet'],
                'download': TORRENT_URL + bangumi['_id'] + '/download.torrent',
                'subtitle_group': bangumi['team_id'],
                'title': bangumi['title'],
                'episode': self.parse_episode(bangumi['title']),
                'time': int(time.mktime(datetime.datetime.strptime(bangumi['publish_time'].split('.')[0],
                                                                   "%Y-%m-%dT%H:%M:%S").timetuple()))
            })

            if os.environ.get('DEBUG'):
                print(ret[index]['download'])

        return ret

    def fetch_bangumi_calendar_and_subtitle_group(self):
        response = get_response(FETCH_URL)
        if not response:
            return []
        bangumi_result, subtitile_result = parser_bangumi(response)
        return bangumi_result, subtitile_result

    def search_by_keyword(self, keyword, count=None):
        if not count:
            count = 3

        rows = []
        result = []

        for i in range(count):
            data = get_response(SEARCH_URL, 'POST', json={'query': keyword, 'p': i + 1})
            if not 'torrents' in data:
                print_warning('No torrents in response data, please re-run')
                return []
            rows.extend(data['torrents'])

        for info in rows:
            if True:
                result.append({
                    'download': TORRENT_URL + info['_id'] + '/download.torrent',
                    'name': keyword,
                    'subtitle_group': info['team_id'],
                    'title': info['title'],
                    'episode': self.parse_episode(info['title']),
                    'time': int(time.mktime(datetime.datetime.strptime(info['publish_time'].split('.')[0],
                                                                       "%Y-%m-%dT%H:%M:%S").timetuple()))
                })

        # Avoid bangumi collection. It's ok but it will waste your traffic and bandwidth.
        result = result[::-1]
        return result



# coding=utf-8
from __future__ import print_function, unicode_literals

import functools
import glob
import gzip
import json
import os
import re
import struct
import sys
import tarfile
import time
from io import BytesIO
from shutil import rmtree, move

import requests
import urllib3
from multiprocessing.pool import ThreadPool

from bgmi import __version__, __admin_version__
from bgmi.config import IS_PYTHON3, BGMI_PATH, DATA_SOURCE, FRONT_STATIC_PATH, SAVE_PATH
from bgmi.lib.constants import SUPPORT_WEBSITE


urllib3.disable_warnings()

# monkey patch for dev
if os.environ.get('DEV', False):  # pragma: no cover
    def replace_url(url):
        if url.startswith('https://'):
            url = url.replace('https://', 'http://localhost:8092/https/')
        elif url.startswith('http://'):
            url = url.replace('http://', 'http://localhost:8092/http/')
        return url


    from requests import request


    def get(url, params=None, **kwargs):
        url = replace_url(url)
        kwargs.setdefault('allow_redirects', True)
        return request('get', url, params=params, **kwargs)


    def post(url, data=None, json=None, **kwargs):
        url = replace_url(url)
        return request('post', url, data=data, json=json, **kwargs)


    requests.get = get
    requests.post = post

if sys.platform.startswith('win'):  # pragma: no cover
    GREEN = ''
    YELLOW = ''
    RED = ''
    COLOR_END = ''
else:
    GREEN = '\033[1;32m'
    YELLOW = '\033[1;33m'
    RED = '\033[1;31m'
    COLOR_END = '\033[0m'

color_map = {
    'print_info': '',
    'print_success': GREEN,
    'print_warning': YELLOW,
    'print_error': RED,
}

indicator_map = {
    'print_info': '[*] ',
    'print_success': '[+] ',
    'print_warning': '[-] ',
    'print_error': '[x] ',
}

NPM_REGISTER_DOMAIN = 'registry.npmjs.com' if os.environ.get('TRAVIS_CI', False) else 'registry.npm.taobao.org'
FRONTEND_NPM_URL = 'https://{}/bgmi-frontend/'.format(NPM_REGISTER_DOMAIN)
PACKAGE_JSON_URL = 'https://{}/bgmi-frontend/{}'.format(NPM_REGISTER_DOMAIN, __admin_version__)


def indicator(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if kwargs.get('indicator', True):
            if IS_PYTHON3:
                func_name = f.__qualname__
            else:
                func_name = f.func_name
            args = (indicator_map.get(func_name, '') + args[0],)
        f(*args, **kwargs)
        sys.stdout.flush()

    return wrapper


def colorize(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if IS_PYTHON3:
            func_name = f.__qualname__
        else:
            func_name = f.func_name
        b, e = color_map.get(func_name, ''), COLOR_END if color_map.get(func_name) else ''
        args = tuple(map(lambda s: b + s + e, args))
        return f(*args, **kwargs)

    return wrapper


@indicator
@colorize
def print_info(message, indicator=True):
    print(message)


@indicator
@colorize
def print_success(message, indicator=True):
    print(message)


@indicator
@colorize
def print_warning(message, indicator=True):
    print(message)


@indicator
@colorize
def print_error(message, exit_=True, indicator=True):
    print(message)
    if exit_:
        exit(1)


def print_version():
    return '''BGmi %sver. %s%s built by %sRicterZ%s with ❤️

Github: https://github.com/BGmi/BGmi
Email: ricterzheng@gmail.com
Blog: https://ricterz.me''' % (YELLOW, __version__, COLOR_END, YELLOW, COLOR_END)


def test_connection():
    try:
        for website in SUPPORT_WEBSITE:
            if DATA_SOURCE == website['id']:
                requests.request('head', website['url'], timeout=10)
    except:
        return False
    return True


def bug_report():  # pragma: no cover
    print_error('It seems that no bangumi found, if https://bangumi.moe can \n'
                '    be opened normally, please submit issue at: https://github.com/BGmi/BGmi/issues',
                exit_=True)


def get_terminal_col():  # pragma: no cover
    # https://gist.github.com/jtriley/1108174
    if not sys.platform.startswith('win'):
        import fcntl
        import termios

        _, col, _, _ = struct.unpack(str('HHHH'), fcntl.ioctl(0, termios.TIOCGWINSZ,
                                                              struct.pack(str('HHHH'), 0, 0, 0, 0)))

        return col
    else:
        try:
            from ctypes import windll, create_string_buffer

            # stdin handle is -10
            # stdout handle is -11
            # stderr handle is -12
            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
            if res:
                (bufx, bufy, curx, cury, wattr,
                 left, top, right, bottom,
                 maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1
                # sizey = bottom - top + 1
                return sizex
            else:
                import subprocess

                cols = int(subprocess.check_output('tput cols'))
                return cols
        except:
            return 80


def check_update(mark=True):
    def update():
        try:
            print_info('Checking update ...')
            version = requests.get('https://pypi.python.org/pypi/bgmi/json',
                                   verify=False).json()['info']['version']

            with open(os.path.join(BGMI_PATH, 'latest'), 'w') as f:
                f.write(version)

            if version > __version__:
                print_warning('Please update bgmi to the latest version {}{}{}.'
                              '\nThen execute `bgmi upgrade` to migrate database'
                              .format(GREEN, version, COLOR_END))
            else:
                print_success('Your BGmi is the latest version.')

            package_json = requests.get(PACKAGE_JSON_URL).json()
            admin_version = package_json['version']
            if glob.glob(os.path.join(FRONT_STATIC_PATH, 'package.json')):
                with open(os.path.join(FRONT_STATIC_PATH, 'package.json'), 'r') as f:
                    local_version = json.loads(f.read())['version']
                if [int(x) for x in admin_version.split('.')] > [int(x) for x in local_version.split('.')]:
                    get_web_admin(method='update')
            else:
                print_info("Use 'bgmi install' to install BGmi frontend / download delegate")
            if not mark:
                update()
                raise SystemExit
        except Exception as e:
            print_warning('Error occurs when checking update, {}'.format(str(e)))

    version_file = os.path.join(BGMI_PATH, 'version')
    if not os.path.exists(version_file):
        with open(version_file, 'w') as f:
            f.write(str(int(time.time())))
        return update()

    with open(version_file, 'r') as f:
        try:
            data = int(f.read())
            if time.time() - 7 * 24 * 3600 > data:
                with open(version_file, 'w') as f:
                    f.write(str(int(time.time())))
                return update()
        except ValueError:
            pass


FETCH_EPISODE_ZH = re.compile("第?\s?(\d{1,3})\s?[話话集]")
FETCH_EPISODE_WITH_BRACKETS = re.compile('[【\[](\d+)\s?(?:END)?[】\]]')
FETCH_EPISODE_ONLY_NUM = re.compile('^([\d]{2,})$')
FETCH_EPISODE_RANGE = re.compile('[\d]{2,}\s?-\s?([\d]{2,})')
FETCH_EPISODE_OVA_OAD = re.compile('([\d]{2,})\s?\((?:OVA|OAD)\)]')
FETCH_EPISODE_WITH_VERSION = re.compile('[【\[](\d+)\s? *v\d(?:END)?[】\]]')
FETCH_EPISODE = (
    FETCH_EPISODE_ZH, FETCH_EPISODE_WITH_BRACKETS, FETCH_EPISODE_ONLY_NUM,
    FETCH_EPISODE_RANGE,
    FETCH_EPISODE_OVA_OAD, FETCH_EPISODE_WITH_VERSION)


def parse_episode(episode_title):
    """
    parse episode from title

    :param episode_title: episode title
    :type episode_title: str
    :return: episode of this title
    :rtype: int
    """

    _ = FETCH_EPISODE_ZH.findall(episode_title)
    if _ and _[0].isdigit():
        return int(_[0])

    _ = FETCH_EPISODE_WITH_BRACKETS.findall(episode_title)
    if _ and _[0].isdigit():
        return int(_[0])

    _ = FETCH_EPISODE_WITH_VERSION.findall(episode_title)
    if _ and _[0].isdigit():
        return int(_[0])

    for split_token in ['【', '[', ' ']:
        for i in episode_title.split(split_token):
            for regexp in FETCH_EPISODE:
                match = regexp.findall(i)
                if match and match[0].isdigit():
                    return int(match[0])
    return 0


def normalize_path(url):
    """
    normalize link to path

    :param url: path or url to normalize
    :type url: str
    :return: normalized path
    :rtype: str
    """
    url = url.replace('http://', 'http/').replace('https://', 'https/')
    illegal_char = [':', '*', '?', '"', '<', '>', '|', "'"]
    for char in illegal_char:
        url = url.replace(char, '')

    if url.startswith('/'):
        return url[1:]
    else:
        return url


def get_web_admin(method):
    # frontend_npm_url = 'https://registry.npmjs.com/bgmi-frontend/'
    print_info('{}ing BGmi frontend'.format(method[0].upper() + method[1:]))

    try:
        r = requests.get(FRONTEND_NPM_URL).json()
        version = requests.get(PACKAGE_JSON_URL).json()
        if 'error' in version and version['reason'] == "document not found":  # pragma: no cover
            print_error("Cnpm has not synchronized the latest version of BGmi-frontend from npm, "
                        "please try it later")
            return
        tar_url = r['versions'][version['version']]['dist']['tarball']
        r = requests.get(tar_url)
    except requests.exceptions.ConnectionError:
        print_warning('failed to download web admin')
        return
    except json.JSONDecodeError:
        print_warning('failed to download web admin')
        return
    admin_zip = BytesIO(r.content)
    with gzip.GzipFile(fileobj=admin_zip) as f:
        tar_file = BytesIO(f.read())

    rmtree(FRONT_STATIC_PATH)
    os.makedirs(FRONT_STATIC_PATH)

    with tarfile.open(fileobj=tar_file) as tar_file_obj:
        tar_file_obj.extractall(path=FRONT_STATIC_PATH)

    for file in os.listdir(os.path.join(FRONT_STATIC_PATH, 'package', 'dist')):
        move(os.path.join(FRONT_STATIC_PATH, 'package', 'dist', file),
             os.path.join(FRONT_STATIC_PATH, file))
    with open(os.path.join(FRONT_STATIC_PATH, 'package.json'), 'w+') as f:
        f.write(json.dumps(version))
    print_success('Web admin page {} successfully. version: {}'.format(method, version['version']))


def convert_cover_url_to_path(cover_url):
    """
    convert bangumi cover to file path

    :param cover_url: bangumi cover path
    :type cover_url:str
    :rtype: str,str
    :return:file_path, dir_path
    """

    cover_url = normalize_path(cover_url)
    file_path = os.path.join(SAVE_PATH, 'cover')
    file_path = os.path.join(file_path, cover_url)
    dir_path = os.path.dirname(file_path)

    return dir_path, file_path


def download_file(url):
    print_info('Download: {}'.format(url))
    return requests.get(url)


def download_cover(cover_url_list):
    """

    :param cover_url_list:
    :type cover_url_list: list
    :return:
    """

    p = ThreadPool(4)
    content_list = p.map(download_file, cover_url_list)
    for index, r in enumerate(content_list):
        dir_path, file_path = convert_cover_url_to_path(cover_url_list[index])
        if not glob.glob(dir_path):
            os.makedirs(dir_path)
        with open(file_path, 'wb') as f:
            f.write(r.content)
    p.close()



# coding=utf-8
from __future__ import print_function, unicode_literals

import glob
import os

from bgmi.config import SAVE_PATH, DOWNLOAD_DELEGATE
from bgmi.lib.models import STATUS_DOWNLOADING, STATUS_NOT_DOWNLOAD, Download
from bgmi.downloader.aria2_rpc import Aria2DownloadRPC
from bgmi.downloader.transmission_rpc import TransmissionRPC
from bgmi.downloader.xunlei import XunleiLixianDownload
from bgmi.utils import print_error


DOWNLOAD_DELEGATE_DICT = {
    'xunlei': XunleiLixianDownload,
    'aria2-rpc': Aria2DownloadRPC,
    'transmission-rpc': TransmissionRPC,
}


def get_download_class(download_obj=None, save_path='', overwrite=True, instance=True):
    if DOWNLOAD_DELEGATE not in DOWNLOAD_DELEGATE_DICT:
        print_error('unexpected delegate {0}'.format(DOWNLOAD_DELEGATE))

    delegate = DOWNLOAD_DELEGATE_DICT.get(DOWNLOAD_DELEGATE)

    if instance:
        delegate = delegate(download_obj=download_obj, overwrite=overwrite, save_path=save_path)

    return delegate


def download_prepare(data):
    """
    list[dict]
    dict[
    name:str, keyword you use when search
    title:str, title of episode
    episode:int, episode of bangumi
    download:str, link to download
    ]
    :param data:
    :return:
    """
    queue = save_to_bangumi_download_queue(data)
    for download in queue:
        # print(download)
        save_path = os.path.join(os.path.join(SAVE_PATH, download.name), str(download.episode))
        if not glob.glob(save_path):
            os.makedirs(save_path)
        # mark as downloading
        download.status = STATUS_DOWNLOADING
        download.save()
        try:
            # start download
            download_class = get_download_class(download_obj=download, save_path=save_path)
            download_class.download()
            download_class.check_download(download.name)

            # mark as downloaded
            download.downloaded()
        except Exception as e:
            if os.getenv('DEBUG'):  # pragma: no cover
                import traceback
                traceback.print_exc()
                raise e

            print_error('Error: {0}'.format(e), exit_=False)
            download.status = STATUS_NOT_DOWNLOAD
            download.save()


def save_to_bangumi_download_queue(data):
    """
    list[dict]
    dict:{
    name;str, keyword you use when search
    title:str, title of episode
    episode:int, episode of bangumi
    download:str, link to download
    }
    :param data:
    :return:
    """
    queue = []
    for i in data:
        download, _ = Download.get_or_create(title=i['title'], download=i['download'],
                                             name=i['name'], episode=i['episode'],
                                             status=STATUS_NOT_DOWNLOAD)

        queue.append(download)

    return queue



# coding=utf-8
from __future__ import print_function, unicode_literals

import os
import re
import time as Time
import urllib

import requests
from bs4 import BeautifulSoup

from bgmi.config import (MAX_PAGE, SHARE_DMHY_URL, IS_PYTHON3)
from bgmi.utils import print_error
from bgmi.website.base import BaseWebsite

if IS_PYTHON3:
    unquote = urllib.parse.unquote
else:
    unquote = urllib.unquote

base_url = SHARE_DMHY_URL


def fetch_url(url, **kwargs):
    ret = None
    try:
        ret = requests.get(url, **kwargs).text
    except requests.ConnectionError:
        print_error('Create connection to {site}... failed'.format(site=SHARE_DMHY_URL), exit_=False)
        print_error(
            'Check internet connection or try to set a DMHY mirror site via: bgmi config SHARE_DMHY_URL <site url>')

    return ret


def parse_bangumi_with_week_days(content, update_time, array_name):
    r = re.compile(array_name + '\.push\(\[\'(.*?)\',\'(.*?)\',\'(.*?)\',\'(.*?)\',\'(.*?)\'\]\)')
    ret = r.findall(content)

    bangumi_list = []
    subtitle_list = []

    for bangumi_row in ret:
        bangumi = {
            "status": 0,
            "subtitle_group": [],
            "name": "",
            "keyword": "",  # bangumi id
            "update_time": "",
            "cover": ""
        }
        (cover_url, name, keyword, subtitle_raw, _) = bangumi_row
        cover = re.findall('(/images/.*)$', cover_url)[0]

        bs = BeautifulSoup(subtitle_raw, 'lxml')
        a_list = bs.find_all('a')

        for a in a_list:
            subtitle_group_name = a.get_text(strip=True)
            subtitle_group_id_raw = re.findall('team_id%3A(.+)$', a['href'])

            if (len(subtitle_group_id_raw) == 0) or subtitle_group_name == '':
                continue

            subtitle_group_id = subtitle_group_id_raw[0]
            # append to subtitle_list
            subtitle_list.append({
                'id': subtitle_group_id,
                'name': subtitle_group_name
            })

            bangumi['subtitle_group'].append(subtitle_group_id)
            pass

        bangumi['name'] = name
        bangumi['update_time'] = update_time
        bangumi['keyword'] = keyword
        bangumi['cover'] = cover

        # append to bangumi_list
        bangumi_list.append(bangumi)

    return bangumi_list, subtitle_list


def parse_subtitle_list(content):
    subtitle_list = []

    bs = BeautifulSoup(content, 'lxml')
    li_list = bs.find_all('li', {'class': 'team-item'})

    for li in li_list:
        subtitle_group_name = li.span.a.get('title')
        subtitle_group_id_raw = re.findall('team_id\/(.+)$', li.span.a.get('href'))

        if (len(subtitle_group_id_raw) == 0) or subtitle_group_name == '':
            continue

        subtitle_group_id = subtitle_group_id_raw[0]
        # append to subtitle_list
        subtitle_list.append({
            'id': subtitle_group_id,
            'name': subtitle_group_name
        })

    return subtitle_list


def unique_subtitle_list(raw_list):
    ret = []
    id_list = list({i['id'] for i in raw_list})
    for row in raw_list:
        if row['id'] in id_list:
            ret.append(row)
            del id_list[id_list.index(row['id'])]

    return ret


class DmhySource(BaseWebsite):
    cover_url = SHARE_DMHY_URL

    def search_by_keyword(self, keyword, count=None):
        """
        return a list of dict with at least 4 key: download, name, title, episode
        example:
        ```
            [
                {
                    'name':"路人女主的养成方法",
                    'download': 'magnet:?xt=urn:btih:what ever',
                    'title': "[澄空学园] 路人女主的养成方法 第12话 MP4 720p  完",
                    'episode': 12
                },
            ]
        ```
        :param keyword: search key word
        :type keyword: str
        :param count: how many page to fetch from website
        :type count: int

        :return: list of episode search result
        :rtype: list[dict]
        """
        if count == None:
            count = 3

        result = []
        search_url = base_url + '/topics/list/'
        for i in range(count):

            params = {'keyword': keyword, 'page': i + 1}

            if os.environ.get('DEBUG', False):  # pragma: no cover
                print(search_url, params)

            r = fetch_url(search_url, params=params)
            bs = BeautifulSoup(r, 'lxml')

            table = bs.find('table', {'id': 'topic_list'})
            if table == None:
                break
            tr_list = table.tbody.find_all('tr', {'class': ''})
            for tr in tr_list:

                td_list = tr.find_all('td')

                if td_list[1].a['class'][0] != 'sort-2':
                    continue

                time_string = td_list[0].span.string
                name = keyword
                title = td_list[2].find('a', {'target': '_blank'}).get_text(strip=True)
                download = td_list[3].a['href']
                episode = self.parse_episode(title)
                time = int(Time.mktime(Time.strptime(time_string, "%Y/%m/%d %H:%M")))

                result.append({
                    'name': name,
                    'title': title,
                    'download': download,
                    'episode': episode,
                    'time': time,
                })

        return result

    def fetch_bangumi_calendar_and_subtitle_group(self):
        """
        return a list of all bangumi and a list of all subtitle group

        list of bangumi dict:
        update time should be one of ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']
        example:
        ```
            [
                {
                    "status": 0,
                    "subtitle_group": [
                        "123",
                        "456"
                    ],
                    "name": "名侦探柯南",
                    "keyword": "1234", #bangumi id
                    "update_time": "Sat",
                    "cover": "data/images/cover1.jpg"
                },
            ]
        ```

        list of subtitle group dict:
        example:
        ```
            [
                {
                    'id': '233',
                    'name': 'bgmi字幕组'
                }
            ]
        ```


        :return: list of bangumi, list of subtitile group
        :rtype: (list[dict], list[dict])
        """
        week_days_mapping = [
            ('Sun', 'sunarray'),
            ('Mon', 'monarray'),
            ('Tue', 'tuearray'),
            ('Wed', 'wedarray'),
            ('Thu', 'thuarray'),
            ('Fri', 'friarray'),
            ('Sat', 'satarray'),
        ]

        bangumi_list = []
        subtitle_list = []

        url = base_url + '/cms/page/name/programme.html'

        r = fetch_url(url)

        for update_time, array_name in week_days_mapping:
            (b_list, s_list) = parse_bangumi_with_week_days(r, update_time, array_name)
            bangumi_list.extend(b_list)
            subtitle_list.extend(s_list)
            pass

        # fetch subtitle
        url = base_url + '/team/navigate/'

        r = fetch_url(url)

        subtitle_list.extend(parse_subtitle_list(r))

        # unique
        subtitle_list = unique_subtitle_list(subtitle_list)

        if os.environ.get('DEBUG', False):  # pragma: no cover
            print(subtitle_list)

        return (bangumi_list, subtitle_list)

    def fetch_episode_of_bangumi(self, bangumi_id, subtitle_list=None, max_page=int(MAX_PAGE)):
        """
        get all episode by bangumi id
        example
        ```
            [
                {
                    "download": "magnet:?xt=urn:btih:e43b3b6b53dd9fd6af1199e112d3c7ff15cab82c",
                    "subtitle_group": "58a9c1c9f5dc363606ab42ec",
                    "title": "【喵萌奶茶屋】★七月新番★[来自深渊/Made in Abyss][07][GB][720P]",
                    "episode": 0,
                    "time": 1503301292
                },
            ]
        ```

        :param bangumi_id: bangumi_id
        :param subtitle_list: list of subtitle group
        :type subtitle_list: list
        :param max_page: how many page you want to crawl if there is no subtitle list
        :type max_page: int
        :return: list of bangumi
        :rtype: list[dict]
        """
        result = []
        keyword = bangumi_id
        search_url = base_url + '/topics/list/'
        for i in range(max_page):

            # params = {'keyword': keyword, 'page': i + 1}

            url = search_url + '?keyword=' + keyword + '&page=' + str(i + 1)

            if os.environ.get('DEBUG', False):  # pragma: no cover
                print(url)

            r = fetch_url(url)
            bs = BeautifulSoup(r, 'lxml')

            table = bs.find('table', {'id': 'topic_list'})
            if table == None:
                break
            tr_list = table.tbody.find_all('tr', {'class': ''})
            for tr in tr_list:

                td_list = tr.find_all('td')

                if td_list[1].a['class'][0] != 'sort-2':
                    continue

                time_string = td_list[0].span.string
                name = keyword
                title = td_list[2].find('a', {'target': '_blank'}).get_text(strip=True)
                download = td_list[3].a['href']
                episode = self.parse_episode(title)
                time = int(Time.mktime(Time.strptime(time_string, "%Y/%m/%d %H:%M")))
                subtitle_group = ''

                tag_list = td_list[2].find_all('span', {'class': 'tag'})

                for tag in tag_list:

                    href = tag.a.get('href')
                    if href == None:
                        continue

                    team_id_raw = re.findall('team_id\/(.*)$', href)
                    if len(team_id_raw) == 0:
                        continue
                    subtitle_group = team_id_raw[0]

                if subtitle_list:
                    if subtitle_group not in subtitle_list:
                        continue

                if os.environ.get('DEBUG', False):  # pragma: no cover
                    print(name, title, subtitle_group, download, episode, time)

                result.append({
                    'name': name,
                    'title': title,
                    'subtitle_group': subtitle_group,
                    'download': download,
                    'episode': episode,
                    'time': time,
                })

        return result

    pass



# coding=utf-8
from __future__ import print_function, unicode_literals

import datetime
import os
import re
import string

from tornado import template

import bgmi.config
from bgmi.lib.fetch import website
from bgmi.lib.download import download_prepare, get_download_class
from bgmi.lib.models import Bangumi, Followed, Filter, Subtitle, STATUS_UPDATED, STATUS_DELETED, STATUS_FOLLOWED
from bgmi.lib.controllers import filter_, source, config, mark, delete, add, search, update, list_
from bgmi.lib.constants import (ACTION_ADD, ACTION_SOURCE, ACTION_DOWNLOAD, ACTION_CONFIG, ACTION_DELETE, ACTION_MARK,
                                ACTION_SEARCH, ACTION_FILTER, ACTION_CAL, ACTION_UPDATE, ACTION_FETCH, ACTION_LIST,
                                DOWNLOAD_CHOICE_LIST_DICT, ACTION_COMPLETE, ACTION_HISTORY,
                                SPACIAL_APPEND_CHARS, SPACIAL_REMOVE_CHARS, SUPPORT_WEBSITE, ACTIONS,
                                actions_and_arguments)
from bgmi.utils import (print_info, print_warning, print_success, print_error,
                        RED, GREEN, YELLOW, COLOR_END, get_terminal_col)
from bgmi.script import ScriptRunner


def source_wrapper(ret):
    result = source(data_source=ret.source)
    globals()["print_{}".format(result['status'])](result['message'])
    return result


def config_wrapper(ret):
    result = config(ret.name, ret.value)
    if (not ret.name) and (not ret.value):
        print(result['message'])
    else:
        globals()["print_{}".format(result['status'])](result['message'])


def search_wrapper(ret):
    data = search(keyword=ret.keyword, count=ret.count, regex=ret.regex_filter)

    for i in data:
        print_success(i['title'])
    if ret.download:
        download_prepare(data)


def mark_wrapper(ret):
    result = mark(name=ret.name, episode=ret.episode)
    globals()["print_{}".format(result['status'])](result['message'])


def delete_wrapper(ret):
    if ret.clear_all:
        delete('', clear_all=ret.clear_all, batch=ret.batch)
    else:
        for bangumi_name in ret.name:
            result = delete(name=bangumi_name)
            globals()["print_{}".format(result['status'])](result['message'])


def add_wrapper(ret):
    for bangumi_name in ret.name:
        result = add(name=bangumi_name, episode=ret.episode)
        globals()["print_{}".format(result['status'])](result['message'])


def list_wrapper(*args):
    result = list_()
    print(result['message'])


def cal_wrapper(ret):
    force_update = ret.force_update
    today = ret.today
    save = not ret.no_save

    runner = ScriptRunner()
    if ret.download_cover:
        cover = runner.get_download_cover()
    else:
        cover = None

    weekly_list = website.bangumi_calendar(
        force_update=force_update, save=save, cover=cover)

    patch_list = runner.get_models_dict()
    for i in patch_list:
        weekly_list[i['update_time'].lower()].append(i)

    def shift(seq, n):
        n %= len(seq)
        return seq[n:] + seq[:n]

    if today:
        weekday_order = (Bangumi.week[datetime.datetime.today().weekday()],)
    else:
        weekday_order = shift(Bangumi.week, datetime.datetime.today().weekday())

    env_columns = 42 if os.environ.get('TRAVIS_CI', False) else get_terminal_col()

    col = 42

    if env_columns < col:
        print_warning('terminal window is too small.')
        env_columns = col

    row = int(env_columns / col if env_columns / col <= 3 else 3)

    def print_line():
        num = col - 3
        split = '-' * num + '   '
        print(split * row)

    for index, weekday in enumerate(weekday_order):
        if weekly_list[weekday.lower()]:
            print(
                '%s%s. %s' % (
                    GREEN, weekday if not today else 'Bangumi Schedule for Today (%s)' % weekday, COLOR_END),
                end='')
            print()
            print_line()
            for i, bangumi in enumerate(weekly_list[weekday.lower()]):
                if bangumi['status'] in (STATUS_UPDATED, STATUS_FOLLOWED) and 'episode' in bangumi:
                    bangumi['name'] = '%s(%d)' % (
                        bangumi['name'], bangumi['episode'])

                half = len(re.findall('[%s]' %
                                      string.printable, bangumi['name']))
                full = (len(bangumi['name']) - half)
                space_count = col - 2 - (full * 2 + half)

                for s in SPACIAL_APPEND_CHARS:
                    if s in bangumi['name']:
                        space_count += bangumi['name'].count(s)

                for s in SPACIAL_REMOVE_CHARS:
                    if s in bangumi['name']:
                        space_count -= bangumi['name'].count(s)

                if bangumi['status'] == STATUS_FOLLOWED:
                    bangumi['name'] = '%s%s%s' % (
                        YELLOW, bangumi['name'], COLOR_END)

                if bangumi['status'] == STATUS_UPDATED:
                    bangumi['name'] = '%s%s%s' % (
                        GREEN, bangumi['name'], COLOR_END)
                try:
                    print(' ' + bangumi['name'], ' ' * space_count, end='')
                except UnicodeEncodeError:
                    continue

                if (i + 1) % row == 0 or i + 1 == len(weekly_list[weekday.lower()]):
                    print()
            print()


def filter_wrapper(ret):
    result = filter_(name=ret.name,
                     subtitle=ret.subtitle,
                     include=ret.include,
                     exclude=ret.exclude,
                     regex=ret.regex)
    if 'data' not in result:
        globals()["print_{}".format(result['status'])](result['message'])
    else:
        print_info('Usable subtitle group: {0}'.format(', '.join(result['data']['subtitle_group'])))
        followed_filter_obj = Filter.get(bangumi_name=ret.name)
        print_filter(followed_filter_obj)
    return result['data']


def update_wrapper(ret):
    update(name=ret.name, download=ret.download, not_ignore=ret.not_ignore)


def download_manager(ret):
    if ret.id:
        # 没有入口..
        download_id = ret.id
        status = ret.status
        if download_id is None or status is None:
            print_error('No id or status specified.')
        # download_obj = NeoDownload.get(_id=download_id)
        # if not download_obj:
        #     print_error('Download object does not exist.')
        # print_info('Download Object <{0} - {1}>, Status: {2}'.format(download_obj.name, download_obj.episode,
        #                                                              download_obj.status))
        # download_obj.status = status
        # download_obj.save()
        print_success('Download status has been marked as {0}'.format(
            DOWNLOAD_CHOICE_LIST_DICT.get(int(status))))
    else:
        status = ret.status
        status = int(status) if status is not None else None
        delegate = get_download_class(instance=False)
        delegate.download_status(status=status)


def fetch_(ret):
    try:
        bangumi_obj = Bangumi.get(name=ret.name)
    except Bangumi.DoesNotExist:
        print_error('Bangumi {0} not exist'.format(ret.name))
        return

    try:
        Followed.get(bangumi_name=bangumi_obj.name)
    except Bangumi.DoesNotExist:
        print_error('Bangumi {0} is not followed'.format(ret.name))
        return

    followed_filter_obj = Filter.get(bangumi_name=ret.name)
    print_filter(followed_filter_obj)

    print_info('Fetch bangumi {0} ...'.format(bangumi_obj.name))
    _, data = website.get_maximum_episode(bangumi_obj, ignore_old_row=False if ret.not_ignore else True)

    if not data:
        print_warning('Nothing.')
    for i in data:
        print_success(i['title'])


def complete(ret):
    # coding=utf-8
    """eval "$(bgmi complete)" to complete bgmi in bash"""
    updating_bangumi_names = [x['name'] for x in Bangumi.get_updating_bangumi(order=False)]

    all_config = [x for x in bgmi.config.__all__ if not x == 'DATA_SOURCE']

    actions_and_opts = {}
    helper = {}
    for action in actions_and_arguments:
        actions_and_opts[action['action']] = [x for x in action.get('arguments', [])
                                              if x['dest'].startswith('-')]
        helper[action['action']] = action.get('help', '')

    if 'bash' in os.getenv('SHELL').lower():  # bash
        template_file_path = os.path.join(os.path.dirname(__file__), '..', 'others', '_bgmi_completion_bash.sh')

    elif 'zsh' in os.getenv('SHELL').lower():  # zsh
        template_file_path = os.path.join(os.path.dirname(__file__), '..', 'others', '_bgmi_completion_zsh.sh')

    else:
        import sys
        print('unsupported shell {}'.format(os.getenv('SHELL').lower()), file=sys.stderr)
        return

    with open(template_file_path, 'r') as template_file:
        shell_template = template.Template(template_file.read(), autoescape='')

    template_with_content = shell_template.generate(actions=ACTIONS,
                                                    bangumi=updating_bangumi_names, config=all_config,
                                                    actions_and_opts=actions_and_opts,
                                                    source=[x['id'] for x in SUPPORT_WEBSITE],
                                                    helper=helper)  # type: bytes

    if os.environ.get('DEBUG', False):  # pragma: no cover
        with open('./_bgmi', 'wb+') as template_file:
            template_file.write(template_with_content)

    template_with_content = template_with_content.decode('utf-8')
    print(template_with_content)


def history(ret):
    m = ('January', 'February', 'March', 'April', 'May', 'June', 'July',
         'August', 'September', 'October', 'November', 'December')
    data = Followed.select(Followed).order_by(Followed.updated_time.asc())
    year = None
    month = None

    print_info('Bangumi Timeline')
    for i in data:
        if i.status == STATUS_DELETED:
            slogan = ' ABANDON'
            color = RED
        else:
            slogan = 'FINISHED'
            color = GREEN

        if not i.updated_time:
            date = datetime.datetime.fromtimestamp(0)
        else:
            date = datetime.datetime.fromtimestamp(int(i.updated_time))

        if date.year != 1970:
            if date.year != year:
                print('%s%s%s' % (GREEN, str(date.year), COLOR_END))
                year = date.year

            if date.year == year and date.month != month :
                print('  |\n  |--- %s%s%s\n  |      |' % (YELLOW, m[date.month - 1], COLOR_END))
                month = date.month

            print('  |      |--- [%s%s%s] %s' % (color, slogan, COLOR_END, i.bangumi_name))


CONTROLLERS_DICT = {
    ACTION_ADD: add_wrapper,
    ACTION_SOURCE: source_wrapper,
    ACTION_DOWNLOAD: download_manager,
    ACTION_CONFIG: config_wrapper,
    ACTION_DELETE: delete_wrapper,
    ACTION_MARK: mark_wrapper,
    ACTION_SEARCH: search_wrapper,
    ACTION_FILTER: filter_wrapper,
    ACTION_CAL: cal_wrapper,
    ACTION_UPDATE: update_wrapper,
    ACTION_FETCH: fetch_,
    ACTION_LIST: list_wrapper,
    ACTION_COMPLETE: complete,
    ACTION_HISTORY: history,
}


def controllers(ret):
    func = CONTROLLERS_DICT.get(ret.action, None)
    if not callable(func):
        return
    else:
        return func(ret)


def print_filter(followed_filter_obj):
    print_info('Followed subtitle group: {0}'.format(', '.join(map(lambda s: s['name'], Subtitle.get_subtitle_by_id(
        followed_filter_obj.subtitle.split(', ')))) if followed_filter_obj.subtitle else 'None'))
    print_info('Include keywords: {0}'.format(followed_filter_obj.include))
    print_info('Exclude keywords: {0}'.format(followed_filter_obj.exclude))
    print_info('Regular expression: {0}'.format(followed_filter_obj.regex))



# coding=utf-8
from __future__ import print_function, unicode_literals
import requests
from bgmi.script import ScriptBase


class Script(ScriptBase):
    class Model(ScriptBase.Model):
        bangumi_name = 'TEST_BANGUMI'
        cover = ''
        update_time = 'Mon'

    def get_download_url(self):
        # fetch and return dict
        resp = requests.get('http://static.ricterz.me/bgmi_test.json').json()

        ret = {}
        for k, v in resp.items():
            ret[int(k)] = v

        return ret


if __name__ == '__main__':  # pragma: no cover
    s = Script()
    s.get_download_url()



#!coding: utf-8
from __future__ import print_function, unicode_literals
import json
import os

import tornado.web

from bgmi import __version__, __admin_version__
from bgmi.config import DANMAKU_API_URL, BGMI_PATH
from bgmi.script import ScriptRunner
from bgmi.utils.utils import normalize_path

COVER_URL = '/bangumi/cover'
WEEK = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')


class BaseHandler(tornado.web.RequestHandler):
    patch_list = None
    latest_version = None

    def get_json(self):
        try:
            return json.loads(self.request.body.decode('utf-8'))
        except (json.JSONDecoder, ValueError):
            return {}

    def jsonify(self, data=None, **kwargs):
        j = {
            'version': __version__,
            'latest_version': self.latest_version,
            'frontend_version': __admin_version__,
            'status': 'success',
            'danmaku_api': DANMAKU_API_URL,
            # 'cover_url': COVER_URL,
            'data': data
        }
        j.update(kwargs)
        self.set_header('content-type', 'application/json; charset=utf-8')
        return json.dumps(j, ensure_ascii=False, indent=2)

    def data_received(self, chunk):
        pass

    def __init__(self, *args, **kwargs):
        if self.latest_version is None:
            if os.path.exists(os.path.join(BGMI_PATH, 'latest')):
                with open(os.path.join(BGMI_PATH, 'latest')) as f:
                    self.latest_version = f.read().strip()

        if self.patch_list is None:
            runner = ScriptRunner()
            self.patch_list = runner.get_models_dict()
            for i in self.patch_list:
                i['cover'] = normalize_path(i['cover'])

        super(BaseHandler, self).__init__(*args, **kwargs)

    def write_error(self, status_code, **kwargs):
        """Override to implement custom error pages.

        ``write_error`` may call `write`, `render`, `set_header`, etc
        to produce output as usual.

        If this error was caused by an uncaught exception (including
        HTTPError), an ``exc_info`` triple will be available as
        ``kwargs["exc_info"]``.  Note that this exception may not be
        the "current" exception for purposes of methods like
        ``sys.exc_info()`` or ``traceback.format_exc``.
        """
        status_code = int(status_code)
        code_and_message_map = {
            400: 'Bad Request',
            401: 'Unauthorized Request',
            # 403: self._reason,
            404: '404 Not Found',
            405: '405 Method Not Allowed',
        }

        self.set_status(status_code)
        self.finish(self.jsonify(status='error',
                                 message=code_and_message_map.get(status_code, self._reason)))



from __future__ import print_function, unicode_literals
import os
from tempfile import NamedTemporaryFile

from bgmi.downloader.base import BaseDownloadService
from bgmi.config import XUNLEI_LX_PATH, BGMI_PATH, TMP_PATH
from bgmi.utils import print_info, print_warning, print_success


class XunleiLixianDownload(BaseDownloadService):
    def __init__(self, *args, **kwargs):
        self.check_delegate_bin_exist(XUNLEI_LX_PATH)
        super(XunleiLixianDownload, self).__init__(*args, **kwargs)

    def download(self):
        print_warning('XunleiLixian is deprecated, please choose aria2-rpc or transmission-rpc.')
        overwrite = '--overwrite' if self.overwrite else ''

        command = [XUNLEI_LX_PATH, 'download', '--torrent', overwrite,
                   '--output-dir={0}'.format(self.save_path), self.torrent,
                   '--verification-code-path={0}'.format(os.path.join(TMP_PATH, 'vcode.jpg'))]

        print_info('Run command {0}'.format(' '.join(command)))
        print_warning('Verification code path: {0}'.format(os.path.join(TMP_PATH, 'vcode.jpg')))
        self.call(command)

    @staticmethod
    def install():
        # install xunlei-lixian
        import tarfile
        import requests
        print_info('Downloading xunlei-lixian from https://github.com/iambus/xunlei-lixian/')
        r = requests.get('https://github.com/iambus/xunlei-lixian/tarball/master', stream=True,
                         headers={'Accept-Encoding': ''})
        f = NamedTemporaryFile(delete=False)

        with f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
        f.close()
        print_success('Download successfully, save at %s, extracting ...' % f.name)
        zip_file = tarfile.open(f.name, 'r:gz')
        zip_file.extractall(os.path.join(BGMI_PATH, 'tools/xunlei-lixian'))
        dir_name = zip_file.getnames()[0]

        print_info('Create link file ...')

        if not os.path.exists(XUNLEI_LX_PATH):
            os.symlink(os.path.join(BGMI_PATH, 'tools/xunlei-lixian/{0}/lixian_cli.py'.format(dir_name)),
                       XUNLEI_LX_PATH)
        else:
            print_warning('{0} already exists'.format(XUNLEI_LX_PATH))

        print_success('All done')
        print_info('Please run command \'{0} config\' to configure your lixian-xunlei '
                   '(Notice: only for Thunder VIP)'.format(XUNLEI_LX_PATH))



# coding=utf-8
from __future__ import print_function, unicode_literals

import glob
import imp
import os
import time
import datetime
import traceback

from bgmi.config import SCRIPT_PATH, MAX_PAGE
from bgmi.lib.download import download_prepare
from bgmi.lib.models import STATUS_UPDATED, STATUS_FOLLOWED
from bgmi.lib.models import Scripts
from bgmi.lib.fetch import DATA_SOURCE_MAP
from bgmi.utils import print_success, print_warning, print_info


class ScriptRunner(object):
    _defined = None
    scripts = []
    download_queue = []

    def __new__(cls, *args, **kwargs):
        if cls._defined is None:

            script_files = glob.glob('{}{}*.py'.format(SCRIPT_PATH, os.path.sep))
            for i in script_files:
                try:
                    s = imp.load_source('script', os.path.join(SCRIPT_PATH, i))
                    script_class = getattr(s, 'Script')()

                    if cls.check(script_class):
                        cls.scripts.append(script_class)
                        print_info('Load script {} successfully.'.format(i))

                except:
                    print_warning('Load script {} failed, ignored'.format(i))
                    if os.getenv('DEBUG_SCRIPT'):  # pragma: no cover
                        traceback.print_exc()
                        # self.scripts = filter(self._check_followed, self.scripts)
                        # self.scripts = filter(self._check_bangumi, self.scripts)

            cls._defined = super(ScriptRunner, cls).__new__(cls, *args, **kwargs)

        return cls._defined

    @classmethod
    def check(cls, script):
        condition = [
            'script.Model().due_date > datetime.datetime.now()',
        ]

        for i in condition:
            try:
                if not eval(i):
                    return False
            except:
                # ignore if error
                if os.getenv('DEBUG_SCRIPT'):  # pragma: no cover
                    traceback.print_exc()

        return True

    def get_model(self, name):
        for script in self.scripts:
            if script.Model.bangumi_name == name:
                return script.Model().obj

    def get_models_dict(self):
        return [dict(script.Model()) for script in self.scripts if script.bangumi_name is not None]

    @staticmethod
    def make_dict(script):
        return [{
            'name': script.bangumi_name,
            'title': '[{}][{}]'.format(script.bangumi_name, k),
            'episode': k,
            'download': v
        } for k, v in script.get_download_url().items()]

    def run(self, return_=True, download=False):
        for script in self.scripts:
            print_info('fetching {} ...'.format(script.bangumi_name))
            download_item = self.make_dict(script)

            script_obj = script.Model().obj

            if not download_item:
                print_info('Got nothing, quit script {}.'.format(script))
                continue

            max_episode = max(download_item, key=lambda d: d['episode'])
            episode = max_episode['episode']
            episode_range = range(script_obj.episode + 1, episode + 1)

            if episode <= script_obj.episode:
                continue

            print_success('{} updated, episode: {}'.format(script.bangumi_name, episode))
            script_obj.episode = episode
            script_obj.status = STATUS_UPDATED
            script_obj.updated_time = int(time.time())
            script_obj.save()

            download_queue = []
            for i in episode_range:
                for e in download_item:
                    if i == e['episode']:
                        download_queue.append(e)

            if return_:
                self.download_queue.extend(download_queue)
                continue

            if download:
                print_success('Start downloading of {}'.format(script))
                download_prepare(download_queue)

        return self.download_queue

    def get_download_cover(self):
        return [script['cover'] for script in self.get_models_dict()]


class ScriptBase(object):
    class Model(object):
        obj = None

        # data
        bangumi_name = None
        cover = None
        update_time = None
        due_date = None

        # source
        source = None

        # fetch_episode_of_bangumi(self, bangumi_id, subtitle_list=None, max_page=MAX_PAGE):
        _bangumi_id = None
        _subtitle_list = []
        _max_page = MAX_PAGE

        def __init__(self):
            if self.bangumi_name is not None:
                s, _ = Scripts.get_or_create(bangumi_name=self.bangumi_name,
                                             defaults={'episode': 0, 'status': STATUS_FOLLOWED})
                self.obj = s

        def __iter__(self):
            for i in ('bangumi_name', 'cover', 'update_time'):
                yield (i, getattr(self, i))

            # patch for cal
            yield ('name', self.bangumi_name)
            yield ('status', self.obj.status)
            yield ('updated_time', self.obj.updated_time)
            yield ('subtitle_group', '')
            yield ('episode', self.obj.episode)

    @property
    def _data(self):
        return {
            'bangumi_id': self.Model._bangumi_id,
            'subtitle_list': self.Model._subtitle_list,
            'max_page': int(self.Model._max_page),
        }

    @property
    def source(self):
        return self.Model.source

    @property
    def name(self):
        return self.Model.bangumi_name

    @property
    def bangumi_name(self):
        return self.Model.bangumi_name

    @property
    def cover(self):
        return self.Model.cover

    @property
    def updated_time(self):
        return self.Model.update_time

    def get_download_url(self):
        """Get the download url, and return a dict of episode and the url.
        Download url also can be magnet link.
        For example:
        ```
            {
                1: 'http://example.com/Bangumi/1/1.mp4'
                2: 'http://example.com/Bangumi/1/2.mp4'
                3: 'http://example.com/Bangumi/1/3.mp4'
            }
        ```
        The keys `1`, `2`, `3` is the episode, the value is the url of bangumi.
        :return: dict
        """
        if self.source is not None:
            source = DATA_SOURCE_MAP.get(self.source, None)()
            if source is None:
                raise Exception('Script data source is invalid, usable sources: {}'
                                .format(', '.join(DATA_SOURCE_MAP.keys())))
            ret = {}
            data = source.fetch_episode_of_bangumi(**self._data)
            for i in data:
                if int(i['episode']) not in data:
                    ret[int(i['episode'])] = i['download']
            return ret
        else:
            return {}


if __name__ == '__main__':
    runner = ScriptRunner()
    runner.run()



# coding=utf-8
from __future__ import print_function, unicode_literals

import bgmi.config
from bgmi.config import BANGUMI_MOE_URL, SHARE_DMHY_URL, unicode_

ACTION_ADD = 'add'
ACTION_FETCH = 'fetch'
ACTION_FILTER = 'filter'
ACTION_DELETE = 'delete'
ACTION_UPDATE = 'update'
ACTION_CAL = 'cal'
ACTION_CONFIG = 'config'
ACTION_DOWNLOAD = 'download'
ACTION_LIST = 'list'
ACTION_MARK = 'mark'
ACTION_SEARCH = 'search'
ACTION_SOURCE = 'source'
ACTION_COMPLETE = 'complete'  # bash completion
ACTIONS = (ACTION_ADD, ACTION_DELETE, ACTION_UPDATE, ACTION_CAL,
           ACTION_CONFIG, ACTION_FILTER, ACTION_FETCH, ACTION_DOWNLOAD,
           ACTION_LIST, ACTION_MARK, ACTION_SEARCH, ACTION_SOURCE, ACTION_COMPLETE)
ACTION_FOLLOWED = 'followed'  # place holder?
ACTION_HISTORY = 'history'

FILTER_CHOICE_TODAY = 'today'
FILTER_CHOICE_ALL = 'all'
FILTER_CHOICE_FOLLOWED = 'followed'
FILTER_CHOICES = (FILTER_CHOICE_ALL, FILTER_CHOICE_FOLLOWED, FILTER_CHOICE_TODAY)

DOWNLOAD_ACTION_LIST = 'list'
DOWNLOAD_ACTION_MARK = 'mark'
DOWNLOAD_ACTION = (DOWNLOAD_ACTION_LIST, DOWNLOAD_ACTION_MARK)

DOWNLOAD_CHOICE_LIST_ALL = 'all'
DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD = 'not_downloaded'
DOWNLOAD_CHOICE_LIST_DOWNLOADING = 'downloading'
DOWNLOAD_CHOICE_LIST_DOWNLOADED = 'downloaded'
DOWNLOAD_CHOICE_LIST_DICT = {
    None: DOWNLOAD_CHOICE_LIST_ALL,
    0: DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD,
    1: DOWNLOAD_CHOICE_LIST_DOWNLOADING,
    2: DOWNLOAD_CHOICE_LIST_DOWNLOADED,
}

DOWNLOAD_CHOICE = (DOWNLOAD_CHOICE_LIST_ALL, DOWNLOAD_CHOICE_LIST_DOWNLOADED,
                   DOWNLOAD_CHOICE_LIST_DOWNLOADING, DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD)

FOLLOWED_ACTION_LIST = 'list'
FOLLOWED_ACTION_MARK = 'mark'
FOLLOWED_CHOICE = (FOLLOWED_ACTION_LIST, FOLLOWED_ACTION_MARK)
SUPPORT_WEBSITE = [
    {
        'view': '萌番组 https://bangumi.moe/',
        'id': "bangumi_moe",
        'url': BANGUMI_MOE_URL
    },
    {
        'view': '蜜柑计划 https://mikanani.me/',
        'id': 'mikan_project',
        'url': 'https://mikanani.me/'
    },
    {
        'view': '动漫花园 http://share.dmhy.org/',
        'id': 'dmhy',
        'url': SHARE_DMHY_URL
    },
]
STATUS_SUCCESS = 'success'
STATUS_WARNING = 'warning'
STATUS_ERROR = 'error'
STATUS_INFO = 'info'

SPACIAL_APPEND_CHARS = ['Ⅱ', 'Ⅲ', '♪', 'Δ', '×', '☆', 'É', '·', '♭', '★']
SPACIAL_REMOVE_CHARS = []

UNSUPPORTED_VIDEO_CODING = ['hevc', ]
COMMON_EXCLUDE_KEYWORD = UNSUPPORTED_VIDEO_CODING

# There should be no `'` in any help message
actions_and_arguments = [
    {
        'action': ACTION_ADD,
        'help': 'Subscribe bangumi.',
        'arguments': [
            {'dest': 'name', 'kwargs': dict(metavar='name', type=unicode_, nargs='+', help='Bangumi name'), },
            {'dest': '--episode',
             'kwargs': dict(metavar='episode', help='Add bangumi and mark it as specified episode.', type=int), },
            {'dest': '--not-ignore', 'kwargs': dict(action='store_true',
                                                    help='Do not ignore the old bangumi detail rows (3 month ago).'), },
        ]
    },
    {
        'action': ACTION_DELETE,
        'help': 'Unsubscribe bangumi.',
        'arguments': [
            {'dest': '--name',
             'kwargs': dict(metavar='name', nargs='+', type=unicode_, help='Bangumi name to unsubscribe.'), },
            {'dest': '--clear-all',
             'kwargs': dict(action='store_true',
                            help='Clear all the subscriptions name will be ignored If you provide this flag.'), },
            {'dest': '--batch',
             'kwargs': dict(action='store_true', help='No confirmation.'), },
        ]
    },
    {
        'action': ACTION_LIST,
        'help': 'List subscribed bangumi.',
    },
    {
        'action': ACTION_FILTER,
        'help': 'Set bangumi fetch filter.',
        'arguments': [
            {'dest': 'name', 'kwargs': dict(metavar='name', type=unicode_, help='Bangumi name to set the filter.'), },
            {'dest': '--subtitle',
             'kwargs': dict(metavar='subtitle', type=unicode_, help='Subtitle group name, split by ",".'), },
            {'dest': '--include', 'kwargs': dict(metavar='include', type=unicode_,
                                                 help='Filter by keywords which in the title, split by ",".'), },
            {'dest': '--exclude', 'kwargs': dict(metavar='exclude', type=unicode_,
                                                 help='Filter by keywords which not int the title, split by ",".'), },
            {'dest': '--regex', 'kwargs': dict(metavar='regex', type=unicode_, help='Filter by regular expression'), },
        ]
    },
    {
        'action': ACTION_UPDATE,
        'help': 'Update bangumi calendar and subscribed bangumi episode.',
        'arguments': [
            {'dest': 'name',
             'kwargs': dict(metavar='name', type=unicode_, nargs='*', help='Update specified bangumi.'), },

            {'dest': '--download',
             'kwargs': dict(action='store', nargs='*', type=int, metavar='episode',
                            help='Download specified episode of the bangumi when updated.'), },

            {'dest': '--not-ignore',
             'kwargs': dict(action='store_true', help='Do not ignore the old bangumi detail rows (3 month ago).')},
        ]
    },
    {
        'action': ACTION_CAL,
        'help': 'Print bangumi calendar.',
        'arguments': [
            {'dest': '--today',
             'kwargs': dict(action='store_true', help='Show bangumi calendar for today.'), },

            {'dest': '--force-update',
             'kwargs': dict(action='store_true', help='Get the newest bangumi calendar from bangumi.moe.'), },

            {'dest': '--download-cover',
             'kwargs': dict(action='store_true', help='Download the cover to local'), },

            {'dest': '--no-save',
             'kwargs': dict(action='store_true', help='Do not save the bangumi data when force update.')},
        ],
    },
    {
        'action': ACTION_CONFIG,
        'help': 'Config BGmi.',
        'arguments': [
            {'dest': 'name',
             'kwargs': dict(nargs='?', type=unicode_, help='Config name',
                            choices=bgmi.config.__all_writable_now__), },

            {'dest': 'value',
             'kwargs': dict(nargs='?', type=unicode_, help='Config value')},
        ],
    },
    {
        'action': ACTION_MARK,
        'help': 'Mark bangumi episode.',
        'arguments': [
            {'dest': 'name',
             'kwargs': dict(type=unicode_, help='Bangumi name'), },

            {'dest': 'episode',
             'kwargs': dict(help='Bangumi episode', type=int), },
        ],
    },
    {
        'action': ACTION_DOWNLOAD,
        'help': 'Download manager.',
        'arguments': [
            {'dest': '--list',
             'kwargs': dict(help='List download queue.', action='store_true'), },
            {'dest': '--mark',
             'kwargs': dict(help='Mark download status with a specific id.', dest='id', type=int), },
            {'dest': '--status',
             'kwargs': dict(type=int, help='Download items status (0: not download, 1: '
                                           'downloading, 2: already downloaded).',
                            choices=[0, 1, 2]), },
        ],
    },
    {
        'action': ACTION_FETCH,
        'help': 'Fetch bangumi.',
        'arguments': [
            {'dest': 'name',
             'kwargs': dict(help='Bangumi name', type=unicode_), },

            {'dest': '--not-ignore',
             'kwargs': dict(action='store_true',
                            help='Do not ignore the old bangumi detail rows (3 month ago).'), },
        ],
    },
    {
        'action': ACTION_SEARCH,
        'help': 'Search torrents from data source by keyword',
        'arguments': [
            {'dest': 'keyword',
             'kwargs': dict(help='Search keyword', type=unicode_), },
            {'dest': '--count',
             'kwargs': dict(type=int, help='The max page count of search result.'), },
            {'dest': '--regex-filter',
             'kwargs': dict(type=unicode_, help='Regular expression filter of title.'), },
            {'dest': '--download',
             'kwargs': dict(action='store_true', help='Download search result.'), },
            {'dest': '--dupe',
             'kwargs': dict(action='store_true',
                            help="Show duplicated episode"), },
        ],
    },
    {
        'action': ACTION_SOURCE,
        'help': 'Select date source bangumi_moe or mikan_project',
        'arguments': [
            {'dest': 'source',
             'kwargs': dict(help='bangumi_moe or mikan_project', type=unicode_,
                            choices=[x['id'] for x in SUPPORT_WEBSITE])},
        ]
    },
    {
        'action': 'install',
        'help': 'Install BGmi front / admin / download delegate'
    },
    {
        'action': 'upgrade',
        'help': 'Check update.'
    },
    {
        'action': 'history',
        'help': 'List your history of following bangumi'
    }

]



# coding: utf-8
from __future__ import print_function, unicode_literals

import datetime
import os
from collections import defaultdict

from icalendar import Calendar, Event
from tornado.web import HTTPError

from bgmi.config import SAVE_PATH
from bgmi.front.base import BaseHandler
from bgmi.lib.models import Download, Followed, Bangumi


class BangumiHandler(BaseHandler):
    def get(self, _):
        if os.environ.get('DEV', False):  # pragma: no cover
            with open(os.path.join(SAVE_PATH, _), 'rb') as f:
                self.write(f.read())
                self.finish()
        else:
            self.set_header('Content-Type', 'text/html')
            self.write('<h1>BGmi HTTP Service</h1>')
            self.write('<pre>Please modify your web server configure file\n'
                       'to server this path to \'%s\'.\n'
                       'e.g.\n\n'
                       '...\n'
                       'autoindex on;\n'
                       'location /bangumi {\n'
                       '    alias %s;\n'
                       '}\n'
                       '...\n\n'
                       'If use want to use Tornado to serve static files, please run\n'
                       '<code>`bgmi config TORNADO_SERVE_STATIC_FILES 1`</code></pre>' % (SAVE_PATH, SAVE_PATH)
                       )
            self.finish()


class RssHandler(BaseHandler):
    def get(self):
        data = Download.get_all_downloads()
        self.set_header('Content-Type', 'text/xml')
        self.render('templates/download.xml', data=data)


class CalendarHandler(BaseHandler):
    def get(self):
        type_ = self.get_argument('type', 0)

        cal = Calendar()
        cal.add('prodid', '-//BGmi Followed Bangumi Calendar//bangumi.ricterz.me//')
        cal.add('version', '2.0')

        data = Followed.get_all_followed()
        data.extend(self.patch_list)

        if type_ == 0:

            bangumi = defaultdict(list)
            [bangumi[Bangumi.week.index(i['update_time']) + 1].append(i['bangumi_name']) for i in data]

            weekday = datetime.datetime.now().weekday()
            for i, k in enumerate(range(weekday, weekday + 7)):
                if k % 7 in bangumi:
                    for v in bangumi[k % 7]:
                        event = Event()
                        event.add('summary', v)
                        event.add('dtstart', datetime.datetime.now().date() + datetime.timedelta(i - 1))
                        event.add('dtend', datetime.datetime.now().date() + datetime.timedelta(i - 1))
                        cal.add_component(event)
        else:
            data = [bangumi for bangumi in data if bangumi['status'] == 2]
            for bangumi in data:
                event = Event()
                event.add('summary', 'Updated: {}'.format(bangumi['bangumi_name']))
                event.add('dtstart', datetime.datetime.now().date())
                event.add('dtend', datetime.datetime.now().date())
                cal.add_component(event)

        cal.add('name', 'Bangumi Calendar')
        cal.add('X-WR-CALNAM', 'Bangumi Calendar')
        cal.add('description', 'Followed Bangumi Calendar')
        cal.add('X-WR-CALDESC', 'Followed Bangumi Calendar')

        self.write(cal.to_ical())
        self.finish()


class NotFoundHandler(BaseHandler):
    def get(self, *args, **kwargs):
        raise HTTPError(404)



# coding=utf-8
from __future__ import print_function, unicode_literals

from collections import defaultdict

import peewee
from peewee import IntegerField, FixedCharField, TextField

from playhouse.shortcuts import model_to_dict

import bgmi.config
from bgmi.config import IS_PYTHON3

if IS_PYTHON3:
    _unicode = str
else:
    input = raw_input
    _unicode = unicode

# bangumi status
STATUS_UPDATING = 0
STATUS_END = 1
BANGUMI_STATUS = (STATUS_UPDATING, STATUS_END)

# subscription status
STATUS_DELETED = 0
STATUS_FOLLOWED = 1
STATUS_UPDATED = 2
FOLLOWED_STATUS = (STATUS_DELETED, STATUS_FOLLOWED, STATUS_UPDATED)

# download status
STATUS_NOT_DOWNLOAD = 0
STATUS_DOWNLOADING = 1
STATUS_DOWNLOADED = 2
DOWNLOAD_STATUS = (STATUS_NOT_DOWNLOAD, STATUS_DOWNLOADING, STATUS_DOWNLOADED)

DoesNotExist = peewee.DoesNotExist

db = peewee.SqliteDatabase(bgmi.config.DB_PATH)


class NeoDB(peewee.Model):
    class Meta:
        database = db


class Bangumi(NeoDB):
    id = IntegerField(primary_key=True)
    name = TextField(unique=True, null=False)
    subtitle_group = TextField(null=False)
    keyword = TextField()
    update_time = FixedCharField(5, null=False)
    cover = TextField()
    status = IntegerField(default=0)

    week = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')

    def __init__(self, **kwargs):
        super(NeoDB, self).__init__(**kwargs)

        update_time = kwargs.get('update_time', '').title()
        if update_time and update_time not in self.week:
            raise ValueError('unexpected update time %s' % update_time)
        self.update_time = update_time
        self.subtitle_group = ', '.join(kwargs.get('subtitle_group', []))

    @classmethod
    def delete_all(cls):
        cls.update(status=STATUS_END).execute()

    @classmethod
    def get_updating_bangumi(cls, status=None, order=True):

        if status is None:
            data = cls.select(cls, Followed.status, Followed.episode) \
                .join(Followed, peewee.JOIN['LEFT_OUTER'], on=(cls.name == Followed.bangumi_name)) \
                .where(cls.status == STATUS_UPDATING).objects()
        else:
            data = cls.select(cls, Followed.status, Followed.episode) \
                .join(Followed, peewee.JOIN['LEFT_OUTER'], on=(cls.name == Followed.bangumi_name)) \
                .where((cls.status == STATUS_UPDATING) & (Followed.status == status)).objects()
        r = []
        for x in data:
            d = model_to_dict(x)
            d['status'] = x.status
            d['episode'] = x.episode
            r.append(d)
        data = r

        if order:
            weekly_list = defaultdict(list)
            for bangumi_item in data:
                weekly_list[bangumi_item['update_time'].lower()].append(dict(bangumi_item))
        else:
            weekly_list = data

        return weekly_list


class Followed(NeoDB):
    bangumi_name = TextField(unique=True)
    episode = IntegerField(null=True)
    status = IntegerField(null=True)
    updated_time = IntegerField(null=True)

    class Meta:
        database = db
        table_name = 'followed'

    @classmethod
    def delete_followed(cls, batch=True):
        q = cls.delete()
        if not batch:
            if not input('[+] are you sure want to CLEAR ALL THE BANGUMI? (y/N): ') == 'y':
                return False

        q.execute()
        return True

    @classmethod
    def get_all_followed(cls, status=STATUS_DELETED, bangumi_status=STATUS_UPDATING):
        join_cond = (Bangumi.name == cls.bangumi_name)
        d = cls.select(cls, Bangumi.name, Bangumi.update_time, Bangumi.cover) \
            .join(Bangumi, peewee.JOIN['LEFT_OUTER'], on=join_cond) \
            .where((cls.status != status) & (Bangumi.status == bangumi_status)) \
            .order_by(cls.updated_time.desc()) \
            .objects()

        r = []
        for x in d:
            dic = dict(**x.__dict__['__data__'])
            dic['update_time'] = x.update_time
            dic['cover'] = x.cover
            r.append(dic)
        return r


class Download(NeoDB):
    name = TextField(null=False)
    title = TextField(null=False)
    episode = IntegerField(default=0)
    download = TextField()
    status = IntegerField(default=0)

    @classmethod
    def get_all_downloads(cls, status=None):
        if status is None:
            data = list(cls.select().order_by(cls.status))
        else:
            data = list(cls.select().where(cls.status == status).order_by(cls.status))

        for index, x in enumerate(data):
            data[index] = model_to_dict(x)
        return data

    def downloaded(self):
        self.status = STATUS_DOWNLOADED
        self.save()


class Filter(NeoDB):
    bangumi_name = TextField(unique=True)
    subtitle = TextField()
    include = TextField()
    exclude = TextField()
    regex = TextField()


class Subtitle(NeoDB):
    id = TextField(primary_key=True, unique=True)
    name = TextField()

    @classmethod
    def get_subtitle_by_id(cls, id_list=None):
        data = list(cls.select().where(cls.id.in_(id_list)))
        for index, subtitle in enumerate(data):
            data[index] = model_to_dict(subtitle)
        return data

    @classmethod
    def get_subtitle_by_name(cls, name_list=None):
        data = list(cls.select().where(cls.name.in_(name_list)))
        for index, subtitle in enumerate(data):
            data[index] = model_to_dict(subtitle)
        return data


script_db = peewee.SqliteDatabase(bgmi.config.SCRIPT_DB_PATH)


class Scripts(peewee.Model):
    bangumi_name = TextField(null=False, unique=True)
    episode = IntegerField(default=0)
    status = IntegerField(default=0)
    updated_time = IntegerField(default=0)

    class Meta:
        database = script_db


def recreate_source_relatively_table():
    table_to_drop = [Bangumi, Followed, Subtitle, Filter, Download]
    for table in table_to_drop:
        table.delete().execute()
    return True


if __name__ == '__main__':  # pragma:no cover
    from pprint import pprint

    d = Bangumi.get_updating_bangumi(status=STATUS_FOLLOWED)
    pprint(d)
